Unit ErrObj;

Interface
Uses Objects;

Type
  StdErrorProc= Procedure( errcode, errinfo: integer );

  PEObject = ^TEObject;
  TEObject = Object( TObject )
    private
    lastErr, lastcode: Word;
    public
    Constructor Init;
    Function GetLastErr: Word;
    Function GetLastCode: Word;
    Procedure Reset;
    Function Failed: Boolean;
    Procedure SetFields(err, info: Word);
    Procedure Error( err, info: Word ); virtual;
    Procedure ForwardError( Const anObj: TEObject );
  End;

Procedure GetGlobalErrorProc( Var proc: StdErrorProc );
Procedure SetGlobalErrorProc( proc: StdErrorProc );

Implementation

Var
  GlobalErrorProc: StdErrorProc;

{+-----------------------------------------------------------------
 | Methods of TEObject
 |
 | TEObject is a base class for all objects that need a way to trap and
 | report errors. It has two fields to store error information; an
 | error id and an error code. Methods are provided to retrieve the
 | contents of the fields. The Error method sets the fields and can
 | be overridden in descendant classes. The default Error method will
 | pass the error info to GlobalErrorProc, which an app can replace 
 | by a custom procedure to trap errors from all objects derived from
 | TEObject. 
 | Unit StdErrs provides such an handler. Error IDs for several common
 | error classes are defined in its header or in deferrs.inc.
 +-----------------------------------------------------------------}
Constructor TEObject.Init;
  Begin
    inherited Init;
    Reset;
  End; { TEObject.Init }

Function TEObject.GetLastErr: Word;
  Begin
    GetLastErr := lastErr;
  End; { TEObject.GetLastErr }

Function TEObject.GetLastCode: Word;
  Begin
    GetLastCode := lastcode;
  End; { TEObject.GetLastCode }

Procedure TEObject.Reset;
  Begin
    lastErr := 0;
    lastcode:= 0;
  End; { TEObject.Reset }

Function TEObject.Failed: Boolean;
  Begin
    Failed := lastErr <> 0;
  End; { TEObject.Failed }

(* SetFields is provided for the benefit of objects that override
   Error but do not want to call the inherited error method. We
   do not use SetFields in this Unit to set the field values (even
   if it would be considered proper by the purists) to avoid the
   overhead of _another_ procedure call. *)
Procedure TEObject.SetFields(err, info: Word);
  Begin
    lastErr := err;
    lastcode := info;
  End; (* Procedure TEObject.SetFields *)

Procedure TEObject.Error( err, info: Word ); 
  Begin
    lastErr := err;
    lastcode := info;
    If err <> 0 Then
      If @GlobalErrorProc <> nil Then
        GlobalErrorProc( err, info );
  End; { TEObject.Error }

Procedure TEObject.ForwardError( Const anObj: TEObject );
  Begin
    SetFields( anObj.GetLastErr, anObj.GetLastCode );
  End; { TEObject.ForwardError }


Procedure GetGlobalErrorProc( Var proc: StdErrorProc );
  Begin
    proc := GlobalErrorProc;
  End;

Procedure SetGlobalErrorProc( proc: StdErrorProc );
  Begin
    GlobalErrorProc := proc; 
  End;



Begin
  GlobalErrorProc := nil;
End.
