{+------------------------------------------------------------
 | Unit FStrClass
 |
 | Version: 1.0  Last modified: 02/08/95, 23:27:09
 | Author : P. Below
 | Project: String Classes
 | Description:
 |   This Unit contains a specialized TStrObj for the handling
 |   of file and pathnames.
 +------------------------------------------------------------}
Unit FStrClass;

Interface

Uses StrClass, Objects;

Const
  PathSeparator = '\';
  NetShareStart : PChar ='\\';
  NetShareMask  : PChar ='\\?*\?*';

  (* error codes for IsValidPath *)
  IVP_InvalidColon = usererror-10;
  IVP_InvalidChar  = usererror-11;
  IVP_NameTooLong  = usererror-12;
  IVP_ExtTooLong   = usererror-13;
  IVP_NameEmpty    = usererror-14;
  IVP_InvalidDrive = usererror-15;
Type
  PFIleName = ^TFilename;
  TFileName = Object( TStrObj )
    Function NewObject: PStrObj; virtual;
    Procedure AppendSeparator;
    Procedure RemoveExtension;
    Procedure RemoveNameAndExtension;
    Procedure RemoveDriveOrNetshare;
    Procedure RemovePath( withDrive: Boolean );
    Procedure ReplaceExtension( newExtension: PChar );
    Function  GetNameAndExtension: Pchar;
    Procedure Expand;
    Procedure ExtractPath( Var pathobj: TStrObj; withDrive: Boolean);
    Function  ExtractDrive: Char;
    Function  ExtractDriveNum: Integer;
    Procedure ExtractName( Var nameObj: TStrObj );
    Procedure ExtractExtension( Var nameObj: TStrObj );
    Procedure ExtractNameAndExtension( Var nameObj: TStrObj );
    Procedure ExtractNetShare( Var nameObj: TStrObj );
    Function  Exists: Boolean;
    Function  IsValidPath: Boolean; virtual;
    Function  IsAbsolutePath: Boolean ;
    Function  HasWildcards: Boolean;
    Function  HasExtension: Boolean ;
    Function  HasFilename: Boolean ;
    Function  HasPath: Boolean ;
    Function  HasDriveletter: Boolean ;
    Function  HasNetshare: Boolean;
    Function  FindEndOfPath: Word;
    Function  FindEndOfName: Word;
    Function  FindStartOfPath: Word ;
    Function  CreatePath: Boolean;
  End;

Const
  RFilename: TStreamRec = (
    ObjType : LastStrObj+1;
    VmtLink : Ofs(Typeof(TFilename)^);
    Load    : @TStrObj.Load;
    Store   : @TStrObj.Store
  );

Implementation

Uses WinDos, Strings, Wintypes, WinProcs, Win31;

Const
{$I deferrs.inc}   (* for usererror *)
{+------------------------------------------------------------------
 | Methods of TFileName
 +------------------------------------------------------------------}
(* object factory *)
Function TFileName.NewObject: PStrObj;
  Begin
    NewObject := New( PFilename, InitBlank );
  End; { TFileName.NewObject }

{************************************************************
 * TFileName.AppendSeparator
 *
 * Parameters:
 *  none
 * Description:
 *  Adds a terminating path separator to the end of the name.
 *  Would turn "D:\temp\local.rdf" into "D:\temp\local.rdf\".
 *  Any whitespace is removed from the ends of the name first.
 *  If the object was empty, we end up with a single backslash!
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 19:17:00 by P. Below
 ************************************************************}
Procedure TFileName.AppendSeparator;
   Var
     len: Word;
   Begin
     Trim;
     len := GetLength;
     If ( len = 0 ) or ( AtChar( len-1 ) <> PathSeparator ) Then
       AppendChar( PathSeparator );
   End; { TFileName.AppendSeparator }

{************************************************************
 * TFileName.RemoveExtension
 *
 * Parameters:
 *  none
 * Description:
 *  Removes the name part after the last dot, plus the dot, from the
 *  name. However, this is only done if no backslash or colon is
 *  found _after_ the dot (which would mean the dot is part of a
 *  directory name or drive spec (which is a _very_ curious drive
 *  indeed!)).
 *  Would turn "D:\temp\local.rdf" into "D:\temp\local"
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 18:54:47 by P. Below
 ************************************************************}
Procedure TFileName.RemoveExtension;
  Var
    dotpos: Word;
  Begin
    dotpos := FindEndOfName;
    If dotpos <> NOT_FOUND Then
      Truncate( dotpos );
  End; { TFileName.RemoveExtension }

{************************************************************
 * TFileName.RemoveNameAndExtension
 *
 * Parameters:
 *  none
 * Description:
 *  This method removes anything from the end of the name up
 *  to and including the last backslash or up to the colon. A colon,
 *  however, is _not_ deleted.
 *  Would turn "D:\temp\local.rdf" into "D:\temp"
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 19:08:35 by P. Below
 ************************************************************}
Procedure TFileName.RemoveNameAndExtension;
  Var
    slashpos : Word;
  Begin
    slashpos := FindEndOfPath;
    If slashpos > 0 Then
      If AtChar( slashpos-1 ) = PathSeparator Then
        Dec( slashpos );
    Truncate( slashpos );
  End; { TFileName.RemoveNameAndExtension }

{************************************************************
 * TFileName.RemoveDriveOrNetshare
 *
 * Parameters:
 *  none
 * Description:
 *  This method deletes anything up to the start of the path,
 *  if present.
 *  Would turn "D:\temp\local.rdf" into "\temp\local.rdf" and
 *  "\\h1h145\auxdisk%below\temp\local.rdf" into "\temp\local.rdf".
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 19:12:40 by P. Below
 ************************************************************}
Procedure TFileName.RemoveDriveOrNetshare;
  Var
    n: Word;
  Begin
    If GetLength > 0 Then Begin
      n := FindStartOfPath;
      If n = NOT_FOUND Then Begin
        (* no path present *)
        n := FindEndOfPath;
      End;
      If n > 0 Then 
        Delete( 0, n );
    End; { If }
  End; { TFileName.RemoveDriveOrNetshare }

{************************************************************
 * TFileName.RemovePath
 *
 * Parameters:
 *  withDrive: if true, the drive letter is removed with the path
 *             if false, only the path is removed
 * Description:
 *  Will remove anything from start of string or first colon to
 *  and including the last path separator. 
 *  Would turn "D:\temp\local.rdf" into "local.rdf" or "D:local.rdf"
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 19:15:33 by P. Below
 ************************************************************}
Procedure TFileName.RemovePath( withDrive: Boolean );
  Var
    pathEnd, pathStart: Word;
  Begin
    If HasPath Then Begin
      pathEnd := FindEndOfPath;
      If pathEnd > 0 Then Begin
        If withDrive Then Begin
          pathStart := 0;
        End { If }
        Else Begin
          pathStart := FindStartOfPath;
          If pathStart = NOT_FOUND Then
            pathStart := 0
          Else
            If HasNetshare Then
              Inc(pathStart);
        End; { Else }
        If pathEnd > pathStart Then
          Delete( pathStart, pathEnd-pathStart );
      End; { If }
    End
    Else
      If withDrive Then
        RemoveDriveOrNetshare;
  End; { TFileName.RemovePath }

{************************************************************
 * TFileName.ReplaceExtension
 *
 * Parameters:
 *  newExtension: pointer to zero-terminated string holding the
 *                new extension. The extension must not contain
 *                whitespace at the start! It can start with a
 *                period, but that is not required.
 * Description:
 *  Deletes the old extension and add the new one, if necessary
 *  with a period prefixed.
 * Error Conditions:
 *
 *
 *Created: 11/30/94 22:24:19 by P. Below
 ************************************************************}
Procedure TFileName.ReplaceExtension( newExtension: PChar );
  Begin
    RemoveExtension;
    If Assigned( newExtension ) Then Begin
      If newExtension^ <> '.' Then
        AppendChar( '.' );
      Append( newExtension );
    End; { if }
  End; { TFileName.ReplaceExtension }

{************************************************************
 * TFileName.GetNameAndExtension
 *
 * Parameters:
 *  none
 * Returns:
 *  a pointer to the start of the filename.ext part of the name.
 *  This is a pointer to the _original_ data, use it only to read
 *  data, never change the data via this pointer!
 * Description:
 *  Uses FindEndOfPath to find the first character past the path
 *  part of the name. If none exists it returns the same as Get.
 *  If the object is empty, you may get Nil.
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 22:27:56 by P. Below
 ************************************************************}
Function  TFileName.GetNameAndExtension: Pchar;
  Begin
    GetNameAndExtension := At( FindEndOfPath );
  End; { TFileName.GetNameAndExtension }

{************************************************************
 * TFileName.Expand
 *
 * Parameters:
 *  none
 * Description:
 *  Uses the WinDOS FileExpand ( DOS Truename ) function on the name
 *  currently in the object.
 * Error Conditions:
 *  may run out of memory if the memory space needs to be expanded.
 *  We allocate space for 255+1 chars, to be able to support long
 *  filenames.
 *
 *Created: 11/30/94 22:31:32 by P. Below
 ************************************************************}
Procedure TFileName.Expand;
  Var
    p: PChar;
  Begin
    If IsValid and Fits( 256, true ) Then Begin
      p := Get;
      FileExpand( p, p );
      CalcLength;
    End;
  End; { TFileName.Expand }

{************************************************************
 * TFileName.ExtractPath
 *
 * Parameters:
 *  pathObj: a string object to copy the path into
 *  withDrive: whether to copy the drive letter with the path
 * Description:
 *  Determines start and end of substring to copy, tries to 
 *  allocate enough space in the target object and copies the 
 *  substring, if successful. if no path is present, the target
 *  object returns empty. Note that the path will end in a backslash
 *  and also start with one, if it does so in the current string!
 * Error Conditions:
 *  May run out of memory, in which case the target object returns
 *  empty.
 *
 *Created: 11/30/94 22:44:00 by P. Below
 ************************************************************}
Procedure TFileName.ExtractPath( Var pathobj: TStrObj; withDrive: Boolean);
  Var
    pathEnd, pathStart, len: Word;
  Begin
    pathObj.DeleteAll;
    If HasPath or withDrive Then Begin
      pathEnd := FindEndOfPath;
      If pathEnd > 0 Then Begin
        If withDrive Then Begin
          pathStart := 0;
        End { If }
        Else Begin
          pathStart := FindStartOfPath;
          If pathStart = NOT_FOUND Then
            pathStart := 0;
        End; { Else }
        len := pathEnd-pathStart;
        If len > 0 Then
          pathObj.PutSubStr( Self, pathStart, len );
      End;
    End; { If }

  End; { TFileName.ExtractPath }

(* return the drive letter, or #0, if the name contains no drive.
   the drive letter is taken to be the letter left of the first
   colon. Case is significant! *)
Function  TFileName.ExtractDrive: Char;
  Var
    colonPos: Word;
  Begin
    ExtractDrive := #0;
    colonPos := Scan( ':' );
    If ( colonPos <> NOT_FOUND ) and ( colonPos > 0 ) Then
      ExtractDrive := AtChar( colonPos-1 );
  End; { TFileName.ExtractDrive }

(* returns the drive number, 
   0: no drive present ( use default ), 
   1..26: A:..Z:, 
   27..52: a:..z: *)
Function  TFileName.ExtractDriveNum: Integer;
  Var
    ch: Char;
  Begin
    ch := ExtractDrive;
    ExtractDriveNum := 0;
    Case ch  Of
      'A'..'Z': ExtractDriveNum := Ord( ch ) - Ord( '@' );
      'a'..'z': ExtractDriveNum := Ord( ch ) - Ord( 'a' )+27;
    End; { Case }
  End; { TFileName.ExtractDriveNum }

{************************************************************
 * TFileName.ExtractName
 *
 * Parameters:
 *  nameObj: initialized object to take the file name
 * Description:
 *  Copies the filename part (without extension) to the passed
 *  object. The filename is anything from the end of the path
 *  (after the last path separator or colon, if such can be found)
 *  up to but not including the last period after the end of path.
 * Error Conditions:
 *  may run out of memory in the target object, in which case that
 *  will be returned empty.
 *
 *Created: 12/02/94 12:14:21 by P. Below
 ************************************************************}
Procedure TFileName.ExtractName( Var nameObj: TStrObj );
  Var
    nameEnd, nameStart, len: Word;
  Begin
    nameObj.DeleteAll;
    nameStart := FindEndOfPath;
    nameEnd := FindEndOfName;
    If ( nameEnd = GetLength ) Then 
      nameObj.Put( At( nameStart ))
    Else Begin
      len := nameEnd-nameStart;
      If len > 0 Then 
        nameobj.PutSubStr( Self, nameStart, len );
    End; { If }
  End; { TFileName.ExtractName }

{************************************************************
 * TFileName.ExtractExtension
 *
 * Parameters:
 *  nameObj: strngobject to take the extension
 * Description:
 *  Copies anything after the last period to nameObj. If the
 *  string is empty or does not contain an extension, the target
 *  object will return empty.
 * Error Conditions:
 *  may run out of memory in the target, in which case it will
 *  return empty.
 *
 *Created: 12/02/94 14:50:28 by P. Below
 ************************************************************}
Procedure TFileName.ExtractExtension( Var nameObj: TStrObj );
  Var
    extStart: Word;
  Begin
    nameObj.DeleteAll;
    extStart := FindEndOfName + 1;
    If extStart < GetLength Then
      nameObj.Put( At( extStart ));
  End; { TFileName.ExtractExtension }

{************************************************************
 * TFileName.ExtractNameAndExtension
 *
 * Parameters:
 *  nameObj: string object to take the name.ext part of this
 *           objects contents.
 * Description:
 *  Copies anything after the end of path (last path separator or
 *  colon) to the target objects. If this object is empty, the 
 *  target will also return empty.
 * Error Conditions:
 *  may run out of memory in the target, in which case it will
 *  return empty.
 *
 *Created: 12/02/94 14:53:02 by P. Below
 ************************************************************}
Procedure TFileName.ExtractNameAndExtension( Var nameObj: TStrObj );
  Begin
    nameObj.Put( At( FindEndOfPath ));
  End; { TFileName.ExtractNameAndExtension }

{************************************************************
 * TFileName.ExtractNetShare
 *
 * Parameters:
 *  nameObj: string object to take the netshare name.
 * Description:
 *  Copies the netshare name, if there is one, to the target 
 *  object. If there is no netshare name, the target will return
 *  empty.
 * Error Conditions:
 *  may run out of memory in the target, in which case it will
 *  return empty.
 *
 *Created: 12/02/94 15:19:16 by P. Below
 ************************************************************}
Procedure TFileName.ExtractNetShare( Var nameObj: TStrObj );
  Var
    len: Word;
  Begin
    nameObj.DeleteAll;
    If HasNetshare Then Begin
      len := FindStartOfPath;
      If len = NOT_FOUND Then
        nameObj.PutObj( Self )
      Else 
        If len > 0 Then 
          nameobj.PutSubStr( Self, 0, len );
    End; { If }
  End; { TFileName.ExtractNetShare }

{************************************************************
 * TFileName.Exists
 *
 * Parameters:
 *  none
 * Returns:
 *  true, if the name in this objects refers to an existing
 *  file or directory, false otherwise. 
 * Description:
 *  Uses FindFirst to try to find the file.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 15:39:43 by P. Below
 ************************************************************}
Function  TFileName.Exists: Boolean;
  Var
    oldState: Word;
    sRec    : TSearchRec;
  Begin
    Exists := False;
    If GetLength > 0 Then Begin
      oldState := SetErrormode( SEM_FAILCRITICALERRORS );
      FindFirst( Get, faAnyFile, sRec );
      Exists := DosError = 0;
      While DosError = 0 Do
        (* complete the search scan. Some network redirectors 
           allegedly become confused if a scan is not continued
           until it returns not found *)
        FindNext( sRec );
      SetErrormode( oldState );
    End; { If }
  End; { TFileName.Exists }

{************************************************************
 * TFileName.IsValidPath
 *
 * Parameters:
 *  none
 * Returns:
 *  true if the contents of this string look like a valid
 *  path+filename, false otherwise
 * Description:
 *  What constitutes a valid filename varies between file systems.
 *  This method assumes the file system to be DOS FAT, with names
 *  limited to 8.3 characters. The name may start with a netshare
 *  or a drive letter. Only one colon may appear and it has to be in 
 *  position 1. Characters <= ' ' are not allowed. An empty string is
 *  not considered a valid name. However, a path ending in a backslash
 *  is accepted. Wildcards are not allowed!
 *  This method is virtual, extend it for other filesystems at your
 *  leisure.
 * Error Conditions:
 *  We report the nature of the problem with error codes, optionally
 *  with the position of the offending character in the error info. 
 *  But this is done bypassing the Error method, so will not be 
 *  reported to the user, if an error handler is installed. 
 *  See Interface for error codes.
 *
 *Created: 12/02/94 15:47:17 by P. Below
 ************************************************************}
Function  TFileName.IsValidPath: Boolean;
  Var
    n: Word;
    ch: Char;
  Function ScanName: Boolean;
    Var
      n: Word;
      len: Word;
      ch: Char;
      p : Pchar;
      state: ( start, netshare, separator, name, ext, finis, err );
    Begin
      p := Get;
      n := 0;
      If HasNetshare Then Begin
        Inc( p, StrLen( NetshareStart ));
        Inc( n, StrLen( NetshareStart ));
        state := netshare;
        len   := 0;
      End { If }
      Else Begin
        If p[ 1 ] = ':' Then Begin
          Inc( p, 2 );
          Inc( n, 2 );
        End;
        state := start;
      End;
      While state < finis Do Begin
        ch := p^;
        (* check if character is valid *)
        If (ch <= ' ') or ( ch='?' ) or ( ch='*' ) Then
          If ch = #0 Then
            state := finis
          Else Begin
            SetFields( IVP_InvalidChar, n );
            state := err;
          End;

        Case state Of
          start:
            Case ch Of
              PathSeparator: state := separator;
              '.': Begin state := ext; len := 0; End
              Else Begin
                state := name;
                len := 0;
              End
            End; { Case }
          netshare:
            If ch = PathSeparator Then Begin
              Inc( len );
              If len > 1 Then
                state:= separator
            End; { Case netshare }
          separator:
            Case ch Of
              PathSeparator: Begin
                  SetFields( IVP_InvalidChar, n );
                  state := err;
                End;
              '.': Begin
                  state := ext;
                  len   := 0;
                End
              Else Begin
                state := name;
                len   := 0;
              End;
            End; { Case }
          name:
            Case ch Of
              PathSeparator: state:= separator;
              '.': Begin
                  state := ext;
                  len   := 0;
                End;
              Else Begin
                Inc( len );
                If len > 8 Then Begin
                  SetFields( IVP_NameTooLong, n );
                  state := err;
                End; { If }
              End;
            End; { Case }
          ext:
            Case ch Of
              PathSeparator: state:= separator;
              '.': Begin
               (* now this is a wee bit of a problem. we may have
                  a parent directory reference here, in which case
                  both before and after we need to have separators *)
                  If (( p[ 1 ] <> PathSeparator ) and ( p[ 1 ] <> #0 )) or
                     (( n=1) or
                      (( AtChar( n-2 ) <> PathSeparator ) and
                       ( AtChar( n-2 ) <> ':' ))) Then
                  Begin
                    SetFields( IVP_InvalidChar, n );
                    state := err ;
                  End
                End;
              Else Begin
                Inc( len );
                If len > 3 Then Begin
                  SetFields( IVP_ExtTooLong, n );
                  state := err;
                End; { If }
              End;
            End; { Case }
        End; { Case }
        Inc( p );
        Inc( n );
      End; { While }
      ScanName :=  state = finis;
    End; { ScanName }
    
  Begin
    IsValidPath := false;
    Trim;
    If GetLength > 0 Then Begin
      (* check 1: colon not allowed if we have a netshare name *)
      If HasNetshare and ( Scan( ':' ) <> NOT_FOUND ) Then Begin
        SetFields( IVP_InvalidColon, Scan( ':' ));
        Exit;
      End;
      (* check 2: if we have a colon, it has to be in position 1 *)
      n:= Scan( ':' );
      If n <> NOT_FOUND Then Begin
        If n <> 1 Then Begin
          SetFields( IVP_InvalidColon, n );
          Exit;
        End { If }
        Else Begin
          (* the drive letter has to be a letter in the range a..zA..Z *)
          ch := UpCase( AtChar( 0 ));
          If ( ch<'A' ) or ( ch>'Z' ) Then Begin
            SetFields( IVP_InvalidDrive, 0 );
            Exit;
          End; { If }
        End; { Else }
      End; { If }
      (* check 3: scan the name from start, looking for invalid characters
                  and name.ext parts that do not conform to the 8.3 format. *)
      IsValidPath := ScanName;
      (* done with all reasonable checks. Of course it is possible to
         contrive names that pass these tests and still are nor valid
         pathnames, e.g. names containing multiple ".." and "." 
         subdir references. Whoever builds such nonsense gets what he
         deserves! *)
    End { If }
    Else
      SetFields( IVP_NameEmpty, 0 );
  End; { TFileName.IsValidPath }

{************************************************************
 * TFileName.IsAbsolutePath
 *
 * Parameters:
 *  none
 * Returns:
 *  true, if the name has an absolute path in it, one that starts
 *  with a path separator at the root level. False otherwise.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 17:23:13 by P. Below
 ************************************************************}
Function  TFileName.IsAbsolutePath: Boolean ;
  Var
    n: Word;
  Begin
    n := FindStartOfPath;
    If n = NOT_FOUND Then
      IsAbsolutePath := False
    Else
      IsAbsolutePath := AtChar( n ) = PathSeparator;
  End; { TFileName.IsAbsolutePath }

{************************************************************
 * TFileName.HasWildcards
 *
 * Parameters:
 *  none
 * Returns:
 *  true, if the filename part contains one of the DOS wildcards
 *  * or ?, false otherwise.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 17:28:07 by P. Below
 ************************************************************}
Function  TFileName.HasWildcards: Boolean;
  Var
    n : Word;
  Begin
    If GetLength = 0 Then 
      HasWildcards := false 
    Else Begin
      n := FindEndOfPath;
      HasWildcards := ( ScanI( '*', n ) <> NOT_FOUND ) or
                      ( ScanI( '?', n ) <> NOT_FOUND );
    End;
  End; { TFileName.HasWildcards }

{************************************************************
 * TFileName.HasExtension
 *
 * Parameters:
 *  none
 * Returns:
 *  true, if the name after the end of path contains a period
 *  followed by something other than a period or #0; false 
 *  otherwise.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 17:38:03 by P. Below
 ************************************************************}
Function  TFileName.HasExtension: Boolean ;
  Var
    n : Word;
    ch: Char;
  Begin
    HasExtension := false;
    If ( GetLength > 0 ) Then Begin
      n := ScanI( '.', FindEndOfPath);
      If n <> NOT_FOUND Then Begin
        ch := AtChar( n+1 );
        HasExtension := ( ch <> #0 ) and ( ch <> '.' );
      End;
    End;
  End; { TFileName.HasExtension }

{************************************************************
 * TFileName.HasFilename
 *
 * Parameters:
 *  none
 * Returns:
 *  true, if the name after the end of path does start with 
 *  something other than a period or #0.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 17:42:29 by P. Below
 ************************************************************}
Function  TFileName.HasFilename: Boolean ;
  Var
    n : Word;
    ch: Char;
  Begin
    If GetLength = 0 Then
      HasFilename := false
    Else Begin
      n := FindEndOfPath;
      ch := AtChar( n );
      HasFilename := ( ch <> #0 ) and ( ch <> '.' );
    End;
  End; { TFileName.HasFilename }

(* HasPAth returns true, if the name includes a path part, which may,
   however, be a single path separator for the root! *)
Function  TFileName.HasPath: Boolean ;
  Begin
    HasPath := FindStartOfPath <> NOT_FOUND;
  End; { TFileName.HasPath }

(* HasDriveLetter returns true, if the name has a drive letter *)
Function  TFileName.HasDriveletter: Boolean ;
  Begin
    HasDriveLetter := ExtractDrive <> #0;
  End; { TFileName.HasDriveletter }

(* HasNetshare returns true, if the name starts with something
   that looks like a netshare *)
Function  TFileName.HasNetshare: Boolean;
  Begin
    HasNetshare := ( Pos( NetshareStart ) = 0 ) and
                   Match( NetShareMask );
  End; { TFileName.HasNetshare }

{************************************************************
 * TFileName.FindEndOfPath
 *
 * Parameters:
 *  none
 * Returns:
 *  Position of the first character after the end of the path
 *  part of the name, counting from 0. This will be the char
 *  immediately after the last backslash or colon. It may be
 *  a null char. If the object is empty, we return 0!
 * Description:
 *  Searches for the last backslash in the name, or, if none is
 *  found, for the last colon. If the name contains no path, the
 *  return is nevertheless the start of the filename part.
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 18:50:26 by P. Below
 ************************************************************}
Function  TFileName.FindEndOfPath: Word;
  Var
    n: Word;
  Begin
    n := RScan( PathSeparator );
    If n = NOT_FOUND Then
      n := RScan( ':' )
    Else Begin
      If (n = 1) and (AtChar( 0 ) = PathSeparator) Then
        (* we found a netshare name, return end position *)
        n := GetLength-1;   (* -1 because it gets incremented below *)
    End;
    If n <> NOT_FOUND Then
      Inc( n )
    Else
      (* we found neither drive letter nor path separator, so anything
         in this string is counted as filename part *)
      n := 0;
    FindEndOfPath := n
  End; { TFileName.FindEndOfPath }

{************************************************************
 * TFileName.FindEndOfName
 *
 * Parameters:
 *  none
 * Returns:
 *  position (counting from 0) of the period separating name and
 *  extension or the position of the terminating 0, if no period
 *  is found in the filename part.
 * Description:
 *  Searches for the last period in the name and checks, whether that
 *  occurs after the last path separator.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 12:21:28 by P. Below
 ************************************************************}
Function  TFileName.FindEndOfName: Word;
  Var
    dotpos: Word;
  Begin
    dotpos := RScan( '.' );
    If dotpos = NOT_FOUND Then
      dotpos := GetLength
    Else 
      If dotpos < FindEndOfPath Then
        dotpos := GetLength;
    FindEndOfName := dotpos;
  End; { TFileName.FindEndOfName }

{************************************************************
 * TFileName.FindStartOfPath
 *
 * Parameters:
 *  none
 * Returns:
 *  position of the first backslash in the path or NOT_FOUND,
 *  if the name contains no path
 * Description:
 *  looks for the first backslash that is not part of a netshare
 *  name.
 * Error Conditions:
 *  none
 *
 *Created: 12/02/94 13:39:09 by P. Below
 ************************************************************}
Function  TFileName.FindStartOfPath: Word ;
  Var
    n: Word;
  Begin
    If HasNetshare Then Begin
      (* we have a netshare name as start. The netshare has the form
         \\servername\servicename
         so we look for the second path separator after the \\ *)
      n:= ScanI( PathSeparator, 2 );
      If n <> NOT_FOUND Then
        n := ScanI( PathSeparator, n+1 );
    End
    Else Begin
      n := Scan( ':' );
      If n <> NOT_FOUND Then Begin
        (* path starts after the colon, but only if we find a
           path separator somewhere after, otherwise we assume no path
           is present, only a filename. *)
        If ScanI( PathSeparator, n ) <> NOT_FOUND Then
          Inc( n )
        Else
          n := NOT_FOUND;
      End
      Else
        (* we have no drive letter, so path starts at position 0, but
           only if we find a path separator somewhere in the string,
           otherwise we assume a naked filename *)
        If Scan( PathSeparator ) <> NOT_FOUND Then
          n := 0
        Else
          n := NOT_FOUND;
    End;
    FindStartOfPath := n;
  End; { TFileName.FindStartOfPath }

{************************************************************
 * TFileName.CreatePath
 *
 * Parameters:
 *  none
 * Returns:
 *  true if all components of the path could be created, false
 *  otherwise.
 * Description:
 *  This method checks whether the path contained in this object
 *  exists and creates any directories it could not find.
 *  The method fail on a path that starts with a netshare name.
 *  it will also remove the filename part first before it actually
 *  starts to check. So if you want all of the contents of this object
 *  to be regarded as a path, add a terminating backslash!
 * Error Conditions:
 *  may run out ot memory used for temp storage or on a create
 *  request.
 *
 *Created: 12/03/94 14:45:31 by P. Below
 ************************************************************}
Function  TFileName.CreatePath: Boolean;
  Var
    tempObj : TFilename;
    n       : Word;
    oldstate: Word;
    p       : PChar;
    sRec    : TSearchRec;
    finished: Boolean;
    errstat : Word;
  Begin
    CreatePath := False;         (* assume error *)
    If GetLength = 0 Then Exit;  (* empty string, how smart! *)
    If HasNetshare Then Exit;    (* no netshare allowed! *)
    tempObj.InitObj( Self );     (* make backup *)
    (* if tempObj.Failed we go ahead anyway, this may result in
       changing this obj but who cares? *)

    { create a full root-relative path }
    Trim;
    Expand;
    If Failed Then Exit;

    { find first path separator }
    n := FindStartOfPath;
    If n = NOT_FOUND Then Exit;  (* no path !!! extremely unlikely but we are
                                    PARANOID! *)

    { enable trapping of critical errors }
    oldState:= SetErrormode( SEM_FAILCRITICALERRORS );
    Repeat
      { find the next path separator after the one we are sitting on now }
      n := ScanI( PathSeparator, n+1 );
      If n <> NOT_FOUND Then Begin
        { replace next path separator with end of string }
        OverwriteChar( #0, n );
        { search for that subpath starting at root }
        FindFirst( Get, faDirectory, sRec );
        finished := DosError = 0 ;
        If finished Then Begin
          { we found something, is it a directory? }
          If (sRec.Attr and faDirectory) = 0 Then Begin
            { found a file! That precents us from creating the directory,
              set access denied error }
            Error( errFilesystem, 5 );
          End
          Else
            { was a directory, so continue scan }
            finished := false;
        End
        Else Begin
          { find failed, so try to create the directory }
          {$I-}
            CreateDir( Get );
            errstat := IOresult;
          {$I+}
            If errstat <> 0 Then Begin
              {create failed!, pass error info to Error and exit loop }
              Error( errFilesystem, errstat );
              finished := True;
            End;
        End;
        { put path separator back }
        OverwriteChar( PathSeparator, n );
      End
      Else Begin
        { we have reached the end of path and encountered no errors }
        CreatePath := True;
        finished := True;
      End;
    Until finished;
    If not tempObj.Failed Then Begin
      (* restore the object to original state before Expand *)
      PutObj( tempObj );
      tempObj.Done;
    End;
    SetErrormode( oldstate );
  End; { TFileName.CreatePath }


End.

