{+------------------------------------------------------------
 | Unit Parser
 |
 | Version: 1.0  Last modified: 02/08/95, 23:28:49
 | Author : P. Below
 | Project: String Classes
 | Description:
 |   This Unit contains a simple string parser object derived from
 |   TStrObj.
 +------------------------------------------------------------}
Unit Parser;

Interface

Uses WinTypes, Objects, StrClass;

Type
  PCharSet   = ^TCharSet;
  TCharSet   = Set of Char;

Const
  DefSeparators: TCharSet = [ #0..' ', ',',';' ];
  DefQuotes    : TCharSet = [ #96, '"' ];
Type
  PStrParser = ^TStrParser;
  TStrParser = Object( TStrObj )
    f_Separators: TCharset; (* set of token separators *)
    f_Token     : TStrObj;  (* object holding the last token found *)
    f_pos       : Word;     (* scan position *)
    Constructor Load( Var S: TStream );
    Destructor Done; virtual;
    Procedure Store( Var S: TStream ); virtual;
    Function NewObject: PStrObj; virtual;
    Function Clone: PStrObj; virtual;
    Procedure SetFieldDefaults; virtual;
    Procedure StartParse; virtual;
    Function  NextToken: PChar;
    Function  GetToken: PChar;
    Function  GetTokenObj: PStrObj;
    Function  GetPos: Word;
    Function  GetCurrChar: Char;
    Function  AtEndOfString: Boolean;
    Procedure Skip;

    Procedure AddSeparator( ch: Char );
    Procedure RemoveSeparator( ch: Char );
    Function  IsSeparator( ch: Char ): Boolean;
    Procedure SetSeparators( Const Seps: TCharSet );
    Procedure GetSeparators( Var Seps: TCharSet );

    Function FindToken( n: Integer ): PChar;
    Function ContainsToken( token: PChar ): Boolean;
    Function NumTokens: Integer;

    (* internal stuff *)
    Procedure SkipSeparators; virtual;
    Procedure FindEndOfToken; virtual;
  End;

  PQStrParser = ^TQStrParser;
  TQStrParser = Object( TStrParser )
    f_quote  : Char;    (* last quote character encountered *)
    Constructor Load( Var S: TStream );
    Procedure Store( Var S: TStream ); virtual;
    Function  NewObject: PStrObj; virtual;
    Function Clone: PStrObj; virtual;
    Procedure SetFieldDefaults; virtual;
    Procedure StartParse; virtual;
    Procedure SkipSeparators; virtual;
    Procedure FindEndOfToken; virtual;
   End;

Const
  RStrParser: TStreamRec =(
    ObjType : LastStrObj+2;    (* +1 already taken by TFilename! *)
    VmtLink : Ofs(Typeof(TStrParser)^);
    Load    : @TStrParser.Load;
    Store   : @TStrParser.Store
  );
  RQStrParser: TStreamRec =(
    ObjType : LastStrObj+3;
    VmtLink : Ofs(Typeof(TQStrParser)^);
    Load    : @TQStrParser.Load;
    Store   : @TQStrParser.Store
  );
  LastParserObj = LastStrObj+3;

Procedure RegisterParsers;

Implementation

Uses Strings;

{************************************************************
 * Procedure RegisterParsers
 *
 * Parameters:
 *  none
 * Description:
 *  registers the objects of this unit and also the TStrObj
 *  from the StrClass unit for streaming.
 * Error Conditions:
 *  none
 *
 *Created: 12/14/94 14:49:13 by P. Below
 ************************************************************}
Procedure RegisterParsers;
  Begin
    RegisterType( RStrObj );
    RegisterType( RStrParser );
    RegisterType( RQStrParser );
  End;

{+------------------------------------------------------------------
 | Methods of TStrParser
 |
 | TStrParser is a simple string parser that can dissect a string
 | into substrings (tokens). Tokens are separated from each other by
 | characters from a set of separators. The default set contains all
 | control characters (Ord(ch) < 32, includes tabs, carriage return,
 | newline and end-of-string marker), the space, comma, and semicolon.
 | You can add other characters to this set or replace the default by
 | your own set of characters.
 |
 | To parse a string you load it into a TStrParser object by one of the
 | methods inherited from TStrObj. Then call StartParse to set the 
 | scan position to the start of string. Repeated calls to NextToken
 | will then yield the tokens in the sequence they appear in the string.
 | Stop scanning when NextToken returns Nil.
 |
 | Note that this object has no understanding of quoted strings! You can
 | add quotes to the set of separators but separators inside quoted strings
 | will _not_ be treated as part of a token!
 |
 | Example:
 |   parser.Init( '"MAIN","Main Window Group", 10' );
 |   parser.StartParse;
 |   p:= parser.NextToken;  -> '"MAIN"'
 |   p:= parser.NextToken;  -> '"Main'
 |   p:= parser.NextToken;  -> 'Window'
 |   p:= parser.NextToken;  -> 'Group"'
 |   p:= parser.NextToken;  -> '10'
 |   p:= parser.NextToken;  -> Nil
 |
 |   parser.AddSeparator( '"' );
 |   parser.StartParse;
 |   p:= parser.NextToken;  -> 'MAIN'
 |   p:= parser.NextToken;  -> 'Main'
 |   p:= parser.NextToken;  -> 'Window'
 |   p:= parser.NextToken;  -> 'Group'
 |   p:= parser.NextToken;  -> '10'
 |
 | Use a TQStrParser object to deal with quoted strings!
 +------------------------------------------------------------------}
(* standard load from stream constructor *)
Constructor TStrParser.Load( Var S: TStream );
  Begin
    inherited Load(S);
    S.Read( f_Separators, SizeOf(f_Separators));
    f_Token.Load(S);
    S.Read( f_Pos, Sizeof(f_Pos));
  End;

(* standard store to stream method *)
Procedure TStrParser.Store( Var S: TStream );
  Begin
    inherited Store(S);
    S.Write( f_Separators, SizeOf(f_Separators));
    f_Token.Store( S );
    S.Write( f_Pos, Sizeof(f_Pos));
  End;

(* object factory, overrides the inherited method *)
Function TStrParser.NewObject: PStrObj;
  Begin
    NewObject := New(PStrParser, InitBlank );
  End;

(* object copy, makes a new object that is a deep copy of Self.
   overrides the inherited method 'cause we have a few new fields
   to copy. *)
Function TStrParser.Clone: PStrObj;
  Var
    p: PStrParser;
  Begin
    p := PStrParser( inherited Clone );
    If Assigned( p ) and not p^.Failed Then Begin
      p^.SetSeparators( f_Separators );
      p^.f_pos := f_pos;
      p^.f_Token.PutObj( f_Token );
    End; { If }
    clone := p;
  End; { TStrParser.Clone }

{************************************************************
 * TStrParser.SetFieldDefaults
 *
 * Parameters:
 *  none
 * Description:
 *  This method is called by all constrcutors to initialize the
 *  fields. Since we have a few new fields we need to override 
 *  TStrObj's method here. We allocate a space of 40 chars for the
 *  token here to reduce probability of running out of memory when
 *  using the object.
 *  The method should _never_ be called outside a constructor
 *  because it will not free any allocated memory!
 * Error Conditions:
 *  may fail due to out of memory conditions when initializing
 *  the f_Token object. Error fields are set accordingly.
 *
 *Created: 12/14/94 15:05:05 by P. Below
 ************************************************************}
Procedure TStrParser.SetFieldDefaults;
  Begin
    inherited SetFieldDefaults;
    f_Separators := DefSeparators;
    f_Token.InitSize( 40 );
    f_pos := 0;
    ForwardError( f_Token );
  End; { TStrParser.SetFieldDefaults }

(* standard destructor *)
Destructor TStrParser.Done;
  Begin
    f_Token.Done;
    inherited Done;
  End; { TStrParser.Done }

(* prepare for a new parsing run by clearing the token string and
   setting the scan position to 0. *)
Procedure TStrParser.StartParse;
  Begin
    f_Token.DeleteAll;
    f_pos := 0;
  End; { TStrParser.StartParse }

{************************************************************
 * TStrParser.NextToken
 *
 * Parameters:
 *  none
 * Returns:
 *  pointer to the string in f_Token, that holds the next token,
 *  or Nil, if there is no next token ( end of string reached ) or
 *  we run out of memory.
 * Description:
 *  This method scans our string for the next token, starting at the
 *  current scan position ( f_pos ). It first calls SkipSeparators to
 *  find the start of the next token, then FindEndOfToken to find its
 *  end. The substring in between is copied to f_Token.
 * Error Conditions:
 *  Has a very rare chance of running out of memory if the token is
 *  long enough to require expansion of f_Token. Since we start out
 *  with a capacitiy of 40 chars, this is rather improbable for 
 *  normal text.
 *
 *Created: 12/14/94 15:29:03 by P. Below
 ************************************************************}
Function  TStrParser.NextToken: PChar;
  Var
    _done: Boolean;
    start: Word;
  Begin
    SkipSeparators;
    f_Token.DeleteAll;
    _done := AtEndOfString;
    start := f_pos;
    NextToken := Nil;
    If not _done Then
      FindEndOfToken;
    If f_pos > start Then Begin
      f_Token.PutSubStr( Self, start, f_pos-start );
      If not f_Token.Failed Then
        NextToken := f_Token.Get;
    End;
  End; { TStrParser.NextToken }

(* return pointer to last token found. Will point to an empty string if
   called before NextToken has been executed at least once! *)
Function  TStrParser.GetToken: PChar;
  Begin
    GetToken := f_Token.Get
  End; { TStrParser.GetToken }

(* return pointer to the TStrObj holding the last token found. Can be
   used to convert that token to a number, for example. The object will
   contain an empty string if this method is called before NextToken has
   been executed at least once. *)
Function  TStrParser.GetTokenObj: PStrObj;
  Begin
    GetTokenObj := @f_Token;
  End; { TStrParser.GetTokenObj }

(* returns current scan position. *)
Function  TStrParser.GetPos: Word;
  Begin
    GetPos := f_pos;
  End; { TStrParser.GetPos }

(* returns true, if the scan position is at the end of string *)
Function  TStrParser.AtEndOfString: Boolean;
  Begin
    AtEndOfString := f_pos >= GetLength;
  End; { TStrParser.AtEndOfString }

(* returns character at current scan position *)
Function TStrParser.GetCurrChar: Char;
  Begin
    GetCurrChar := AtChar( f_pos );
  End; { TStrParser.GetCurrChar }

(* advance the scan position by one, if not already at end of string,
   thus skipping the next character. *)
Procedure TStrParser.Skip; 
  Begin
    If not AtEndOfString Then 
      Inc( f_pos );
  End; { TStrParser.Skip }

(* add the passed character to the set of separators *)
Procedure TStrParser.AddSeparator( ch: Char );
  Begin
    Include( f_Separators, ch );
  End; { TStrParser.AddSeparator }

(* remove the passed character from the set of separators. Will refuse
   to remove #0, since we need that to detect the end of string. *)
Procedure TStrParser.RemoveSeparator( ch: Char );
  Begin
    (* don't allow the end of string marker to be removed! *)
    If ch <> #0 Then
      Exclude( f_Separators, ch );
  End; { TStrParser.RemoveSeparator }

(* returns true, if the passed character is in the set of separators,
   false otherwise. *)
Function  TStrParser.IsSeparator( ch: Char ): Boolean;
  Begin
    IsSeparator := ch In f_Separators
  End; { TStrParser.IsSeparator }

(* replace the set of separators with the passed set. The #0 character
   is always added to the resulting set because we need it to detect the
   end of string. *)
Procedure TStrParser.SetSeparators( Const Seps: TCharSet );
  Begin
    f_Separators := Seps;
    Include( f_Separators, #0 );
  End; { TStrParser.SetSeparators }

(* copies the current set of separators to the passed variable. *)
Procedure TStrParser.GetSeparators( Var Seps: TCharSet );
  Begin
    Seps := f_Separators
  End; { TStrParser.GetSeparators }

(* returns a pointer to the n-th token ( counting from 1 ) in the 
   current string, or nil, if there is no n-th token. This method 
   starts a new scan of the string, so do not call it in a loop that
   calls NextToken! *)
Function TStrParser.FindToken( n: Integer ): PChar;
  Var
    p: Pchar;
    i: Integer;
  Begin
    StartParse;
    For i:= 1 To n Do Begin
      p := NextToken;
      If AtEndOfString Then Break;
    End; { For }
    FindToken := p;
  End; { TStrParser.FindToken }

(* returns True, if the string contains the passed token. The comparison
   is case-sensitive! Note that we return False if the passed string is
   only a substring of one of the tokens, we only honor identity! *)
Function TStrParser.ContainsToken( token: PChar ): Boolean;
  Var
    p: PChar;
  Begin
    ContainsToken := FALSE;
    StartParse;
    Repeat
      p := NextToken;
      If Assigned( p ) Then
        If StrComp( p, token ) = 0 Then Begin
          ContainsToken := True;
          Break;
        End; { If }
    Until p = Nil;
  End; { TStrParser.ContainsToken }

(* return the number of tokens in this string *)
Function TStrParser.NumTokens: Integer;
  Var
    n: Integer;
  Begin
    StartParse;
    n := 0;
    While NextToken <> Nil Do
      Inc( n );
    NumTokens := n;
  End; { TStrParser.NumTokens }

(* internal stuff *)
{************************************************************
 * TStrParser.FindEndOfToken
 *
 * Parameters:
 *  none
 * Description:
 *  This method is called by NextToken. The scan position is on
 *  the first character of a token. We advance the scan position
 *  until a separator is found. The end-of-string marker ( #0 ) is
 *  always in the separator set.
 *  Override this method to implement different end-of-token conditions.
 * Error Conditions:
 *  none
 *
 *Created: 12/14/94 15:26:09 by P. Below
 ************************************************************}
Procedure TStrParser.FindEndOfToken;
  Var
    found: Boolean;
    ch   : Char;
  Begin
    found := False;
    While not found Do Begin
      ch := AtChar( f_pos );
      If ch IN f_Separators Then
        found := true
      Else
        Inc( f_pos )
    End; { While }
  End; { TStrParser.FindEndOfToken }

{************************************************************
 * TStrParser.SkipSeparators
 *
 * Parameters:
 *  none
 * Description:
 *  Called from NextToken to find the start of the next token.
 *  Advances the current scan position as long as the character
 *  at that position is in the set of separators. Will stop at
 *  the end of string. After SkipSeparators the current scan 
 *  position will either be on the first character of the next
 *  token or on the end-of-string marker.
 *  Override this method to implement a different way to find the
 *  start of the next token.
 * Error Conditions:
 *  none
 *
 *Created: 12/14/94 15:46:05 by P. Below
 ************************************************************}
Procedure TStrParser.SkipSeparators;
  Var
    ch: Char;
  Begin
    Repeat
      ch := AtChar( f_pos );
      If ( ch In f_Separators ) and ( ch <> #0 ) Then
        Inc( f_pos )
      Else
        Break;
    Until False;
  End; { TStrParser.SkipSeparators }

{+------------------------------------------------------------------
 | Methods of TQStrParser
 |
 | TQStrParser is an extended version of TStrParser, that knows how
 | to deal with quoted strings. It recognized both single and double
 | quotes (" = #34, ' = #96). When it finds a starting quote, it will
 | treat the whole text up to the next matching quote as a token, even
 | if it contains separators. Otherwise, operations are as for TStrParser.
 |
 | Example:
 |   parser.Init( '"MAIN","Main Window Group", 10' );
 |   parser.StartParse;
 |   p:= parser.NextToken;  -> 'MAIN'
 |   p:= parser.NextToken;  -> 'Main Window Group'
 |   p:= parser.NextToken;  -> '10'
 |   p:= parser.NextToken;  -> Nil
 +------------------------------------------------------------------}
Constructor TQStrParser.Load( Var S: TStream );
  Begin
    inherited Load(S);
    S.Read( f_Quote, SizeOf(f_Quote));
  End;

Procedure TQStrParser.Store( Var S: TStream );
  Begin
    inherited Store(S);
    S.Write( f_Quote, SizeOf(f_Quote));
  End;

Function TQStrParser.NewObject: PStrObj;
  Begin
    NewObject := New(PQStrParser, InitBlank );
  End;

(* object copy, makes a new object that is a deep copy of Self.
   overrides the inherited method 'cause we have a few new fields
   to copy. *)
Function TQStrParser.Clone: PStrObj;
  Var
    p: PQStrParser;
  Begin
    p := PQStrParser( inherited Clone );
    If Assigned( p ) and not p^.Failed Then Begin
      p^.f_quote := f_quote;
    End; { If }
    clone := p;
  End; { TQStrParser.Clone }

Procedure TQStrParser.SetFieldDefaults;
  Begin
    inherited SetFieldDefaults;
    f_quote := #0;
  End; { TStrParser.InitBlank }

Procedure TQStrParser.StartParse;
  Begin
    inherited StartParse;
    f_quote := #0;
  End; { TQStrParser.StartParse }

{************************************************************
 * TQStrParser.FindEndOfToken
 *
 * Parameters:
 *  none
 * Description:
 *  This method is called by NextToken to find the end of the 
 *  current token.
 *  If the last SkipSeparators call found a quote, this method 
 *  will skip to the matching closing quote, otherwise it skips
 *  to the next separator via the inherited method.
 * Error Conditions:
 *  none
 *
 *Created: 12/14/94 16:12:01 by P. Below
 ************************************************************}
Procedure TQStrParser.FindEndOfToken;
  Var
    found: Boolean;
    ch   : Char;
  Begin
    found := False;
    If f_quote = #0 Then
      inherited FindEndOfToken
    Else
      While not found Do Begin
        ch := AtChar( f_pos );
        If ( ch <> f_quote ) and ( ch <> #0 )Then
          Inc( f_pos )
        Else
          found := True
      End; { While }
  End; { TQStrParser.FindEndOfToken }

{************************************************************
 * TQStrParser.SkipSeparators
 *
 * Parameters:
 *  none
 * Description:
 *  This method is called by NextToken to find the start of the
 *  next token. 
 *  If the last token was a quoted string, the current scan position
 *  is on the closing quote of that string, so we advance beyond that.
 *  The scan position is then advanced until we find a quote, which is
 *  stored in the f_quote field to allow FindEndOfToken to find its
 *  partner, or until we find a character not in the set of separators
 *  or the end of string.
 *  On exit the scanning position will be on the start of a token or
 *  on the end-of-string marker.
 * Error Conditions:
 *  none
 *
 *Created: 12/14/94 16:14:59 by P. Below
 ************************************************************}
Procedure TQStrParser.SkipSeparators;
  Var
    ch: Char;
  Begin
    If f_quote <> #0 Then Begin
      Inc( f_pos );
      f_quote := #0;
    End;
    Repeat
      ch := AtChar( f_pos );
      If ch In DefQuotes Then Begin
        f_quote := ch;
        Inc( f_pos );
        Break;
      End;
      If ( ch In f_Separators ) and ( ch <> #0 ) Then
        Inc( f_pos )
      Else
        Break;
    Until False;
  End; { TQStrParser.SkipSeparators }

End.


