{+------------------------------------------------------------
 | Unit StrClass
 |
 | Version: 1.0  Last modified: 02/08/95, 23:16:00
 | Author : P. Below
 | Project: Tools
 | Description:
 |   This Unit provides an object that makes working with zero-
 |   terminated strings safer and easier. It hides the details 
 |   of memory handling and can be used both like a PChar or
 |   a Pascal string. It also provides all the basic string 
 |   manipulation methods found in BPs System and Strings Unit,
 |   plus a couple more. 
 |
 |   A special debugging version is incorporated. It provides
 |   enhanced checking on memory overwrites and detailed error
 |   messages. The non-debug version relies on the default error
 |   handling mechanism provided by TEObject, from which TStrObj
 |   descends. The only error in that version is out of memory.
 |   The debugging version will check on every one of the Getxxx
 |   methods and will slow down performance significantly!
 +------------------------------------------------------------}
Unit StrClass;

Interface

Uses Objects, WinTypes, ErrObj;

(* disable the following conditional if you do not want to get the
   debugging version automatically if the $D+ switch is set for
   compilation! *)
{$IFOPT D+}
  {$DEFINE DEBUG}
{$ENDIF}
Const
{$I deferrs.inc}      (* defines usererror *)
  SOERR_ConversionFailed = usererror;  (* this is nonfatal *)
  SOERR_BlockTooLarge = usererror -1;
  SOERR_MemOverwrite = usererror - 2;  (* only used in debug version *)
  SOERR_NotZeroterminated = usererror - 3; (* ditto *)
  SOERR_SelfReference = usererror - 4;
{$IFDEF DEBUG}
  SO_Extrabytes = 4;  (* allocate 4 bytes in excess of string length *)
  SO_Guard = $007C;    (*  : pChar = '|'  *)
{$ELSE}
  SO_Extrabytes = 2;  (* allocate 2 bytes in excess of string length *)
{$ENDIF}

  NOT_FOUND = High( Word );
Type
  TLen = LongInt;

  PStrObj = ^TStrObj;
  TStrObj = Object( TEObject )
  private
    buf  : PChar;
    size : Word;
  public
    (* Naming conventions for methods: 
       Methods are grouped according to function, the group
       name serves as a prefix for the method name. The name
       may have a suffix, which denotes the data type the method
       operates on or returns. A method without suffix (name=group
       name) operates on or returns PChars, which should be the
       most prevalent use under Windows. "String" as a data type
       refers to Pascal strings! *)

    (* constructors, destructor, object factory and stream support *)
    Constructor Init( aStr: Pchar );
    Constructor InitBlank;
    Constructor InitString( Const aStr: String );
    Constructor InitObj( Const aStrObj: TStrObj );
    Constructor InitSize( len: Word );
    Constructor InitResourceString( hInst: THandle; id: Word );
    Constructor Load( Var S: TStream );
    Destructor Done; virtual;
    Procedure Store( Var S: TStream ); virtual;
    Function Clone: PStrObj; virtual;
    Function NewObject: PStrObj; virtual;
    Procedure SetFieldDefaults; virtual;


{$IFDEF DEBUG}
    (* debugging helper methods *)
    Procedure Error( err, info: Word ); virtual;
    Procedure Validate;
{$ENDIF}

    (* field access methods, read *)
    Function GetLength: Word;
    Function GetMemSize: Word;
    Function GetCapacity: Word;
    Function GetBuffer: PChar;
    Function Get: PChar;
    Function GetString: String;
    Function GetPString: PString;
    Function IsValid: Boolean;

    Function SubStr( aBuf: PChar; bufsize, start, count: Word ): PChar;
    Function SubStrString( start, count: Word ): String;
    Function SubStrObj( start, count: Word ): PStrObj;

    (* methods to put data into the object *)
    Procedure Put( aStr: PChar );
    Procedure PutChar( c: Char );
    Procedure PutString( Const aStr: string );
    Procedure PutInteger( aValue: LongInt; fieldwidth: Integer );
    Procedure PutReal( aValue: Real; fieldwidth, precision: Integer );
    Procedure PutObj( Const aStrObj: TStrObj );
    Procedure PutFormat( fmtStr: PChar; Var data );
    Procedure PutFill( count: Word; c: Char );
    Procedure PutResourceString( inst: THandle; id: Word );
    Procedure PutSubstr( Const aStrObj: TStrObj; startAt, count: Word );

    (* methods to change existing data.   *)
    Procedure Append( aStr: PChar );
    Procedure AppendChar( c: Char );
    Procedure AppendEOL; virtual;
    Procedure AppendString( Const aStr: String );
    Procedure AppendInteger( aValue: LongInt; fieldwidth: Integer );
    Procedure AppendReal( aValue: Real; fieldwidth, precision: Integer );
    Procedure AppendObj( Const aStrObj: TStrObj );
    Procedure AppendFormat( fmtStr: PChar; Var data );
    Procedure AppendFill( count: Word; c: Char );
    Procedure AppendPad( withC: Char; toLength: Word );
    Procedure AppendResourceString( inst: THandle; id: Word );

    Procedure Insert( aStr: Pchar; atPos: Word );
    Procedure InsertChar( c: char; atPos: Word );
    Procedure InsertString( Const aStr: String; atPos: Word );
    Procedure InsertInteger( aValue: LongInt; fieldwidth: Integer ; atPos: Word);
    Procedure InsertReal( aValue: Real; fieldwidth, precision: Integer ; atPos: Word);
    Procedure InsertObj( Const aStrObj: TStrObj; atPos: Word ); virtual;
    Procedure InsertFormat( fmtStr: PChar; Var data; atPos: Word );
    Procedure InsertFill( count: Word; c: Char; atPos: Word );
    Procedure InsertResourceString( inst: THandle; id: Word; atPos: Word );

    Procedure Delete( atPos, count: Word ); 
    Procedure DeleteAll;
    Procedure Truncate( atPos: Word );
    Procedure RTrim;
    Procedure LTrim;
    Procedure Trim;

    Function Compare( aStr: PChar; caseSensitive: Boolean ): Integer;
    Function CompareString( Const aStr: String; caseSensitive: Boolean): Integer;
    Function CompareObj( Const aStrObj: TStrObj; caseSensitive: Boolean): Integer;

    Function Replace( aPattern, aStr: PChar ): Boolean;
    Function ReplaceString( Const aPattern, aStr: String ): Boolean;

    Procedure Overwrite( aStr: PChar; atPos: Word );
    Procedure OverwriteChar( c: char; atPos: Word );
    Procedure OverwriteString( Const aStr: String; atPos: Word );
    Procedure OverwriteFill( count: Word; c: Char; atPos: Word );
    (* no further verions provided here since this group should see
       much less usage than the Put, Append, and Insert groups.
       All missing operations can easily be accomplished using 
       temp objects and Overwrite. *)

    (* conversion methods *)
    Procedure ToUpper;
    Procedure ToLower;
    Procedure ToOEM;
    Procedure ToANSI;
    (* the following two methods report errors via the Failed method
       inherited from TEObject. A conversion error will NOT trigger
       the Error method! *)
    Function  ToInteger: LongInt;
    Function  ToReal: Real;

    (* methods to search the data *)
    Function Match( anExpression: PChar ): Boolean;
    Function MatchString( Const anExpression: String ): Boolean;

    Function Pos( aStr: PChar ): Word;
    Function PosString( Const aStr: String ): Word;
    Function PosObj( Const aStrObj: TStrObj ): Word;

    Function At( index: Word ): PChar;
    Function AtChar( index: Word ): Char ;
    Function AtString( index: Word ): Char;

    Function Scan( c: Char ): Word;
    Function ScanString( c: Char ):Word;
    Function RScan( c: Char ): Word;
    Function RScanString( c: Char ): Word;

    (* methods to search starting at arbitrary position inside the 
       string. These are only provided for PChars. The I in the name
       stands for incremental *)
    Function PosI( aStr: PChar; startAt: Word ): Word;
    Function ScanI( c: Char; startAt: Word ): Word;
    Function RScanI( c: Char; startAt: Word ): Word;
    
    (* memory handling methods, considered for internal use only,
       only exposed for the benefit of derived objects! *)
    Function  Alloc( newSize: Word ): Boolean; 
    Procedure DisposeMem;
    Procedure ReAlloc( newsize: Word; preserve: Boolean );
    Procedure Compact;
    Procedure SetLength( len: Word );
    Procedure CalcLength;
    Function  Fits( neededSize: TLen; preserve: Boolean  ): Boolean;
  End;

  PFStrObj = ^TFStrObj;
  TFStrObj = Object( TStrObj )
    Constructor InitReadLn( Var f: Text );
    Constructor InitRead( Var f: Text; count: Word );
    Function NewObject: PStrObj; virtual;
    Procedure FWriteLn( Var f: Text );
    Procedure FWrite( Var f: Text );
  End;

  PMsgObj = ^TMsgObj;
  TMsgObj = Object( TStrObj )
    Function Display( parentHwnd: Hwnd; title: PChar; msgboxFlags: Word ):
      Word;
    Function DisplayWithID( parentHwnd: Hwnd; resID, msgboxFlags: Word ):
      Word;
    Function NewObject: PStrObj; virtual;
  End;

Const
  RStrObj : TStreamRec = (
    ObjType : 2543;
    VmtLink : Ofs(Typeof(TStrObj)^);
    Load    : @TStrObj.Load;
    Store   : @TStrObj.Store
  );
  RFStrObj : TStreamRec = (
    ObjType : 2544;
    VmtLink : Ofs(Typeof(TFStrObj)^);
    Load    : @TStrObj.Load;
    Store   : @TStrObj.Store
  );
  RMsgObj : TStreamRec = (
    ObjType : 2545;
    VmtLink : Ofs(Typeof(TMsgObj)^);
    Load    : @TStrObj.Load;
    Store   : @TStrObj.Store
  );
  LastStrObj = 2545;

Procedure RegisterStrObjects;

Implementation

Uses Strings,
{$IFDEF WINDOWS }
WinProcs, OWindows
{$ENDIF}
;

Procedure RegisterStrObjects;
  Begin
    RegisterType( RStrObj );
    RegisterType( RFStrObj );
    RegisterType( RMsgObj );
  End;

{+------------------------------------------------------------------
 | Description of TStrObj
 |
 | The object stores its data in a buffer allocated on the heap. The
 | layout of that buffer is as follows:
 |
 | length byte, characters, terminating #0, guard word
 |
 | The guard word is only present in the debugging version. The buffer
 | can hold up to 65533 characters (65531 for the debug version). For
 | any string longer than 255 chars, the length byte will remain set
 | to 255; all characters above that limit cannot be accessed as Pascal
 | strings.
 |
 | The buffer will grow up to the maximal size, if required, but it
 | will not automatically shrink to fit smaller string sizes. This
 | prevents some of the overhead of using the heap manager and also cuts
 | down on heap fragmentation. You have to call the Compact method to
 | shrink the memory down to what's actually needed.
 |
 | For descriptions of the methods, see the methods header.
 | Note that there is no method to hook an externally created buffer to
 | this object. This is on purpose because it would interfere with the
 | assumptions this object makes about the organization of the buffer!
 +------------------------------------------------------------------}

{+------------------------------------------------------------------
 | Methods of TStrObj
 +------------------------------------------------------------------}
(* constructors, destructor, object factory and stream support *)

(* The Constructors differ only in the kind of data they take as
   parameter. They all call the InitBlank Constructor first and
   then the Put method for the parameter type. The finished product
   is an objects that holds a _copy_ of the data passed to the 
   Constructor. The size is just enough to hold the data. *)
Procedure TStrObj.SetFieldDefaults;
  Begin
    buf := Nil;
    size:= 0;
  End;
Constructor TStrObj.InitBlank;
  Begin
    inherited Init;
    SetFieldDefaults;
  End; { TStrObj.InitBlank }

Constructor TStrObj.Init( aStr: Pchar ); 
  Begin
    inherited Init;
    SetFieldDefaults;
    Put( aStr );
  End; { TStrObj.Init }

Constructor TStrObj.InitString( Const aStr: String );
  Begin
    inherited Init;
    SetFieldDefaults;
    PutString( aStr );
  End; { TStrObj.InitString }

Constructor TStrObj.InitObj( Const aStrObj: TStrObj );
  Begin
    If @aStrObj <> @Self Then Begin
      inherited Init;
      SetFieldDefaults;
      Put( aStrObj.Get );
    End
    Else
      Error( SOERR_SELFREFERENCE, 0 );
  End; { TStrObj.InitObj }

(* The InitSize Constructor is slightly different, i takes the number
   of characters the object should be able to hold as parameter.
   This is the needed capacity of the object, the amount of memory
   allocated will be larger!. The memory is filled with #0. *)
Constructor TStrObj.InitSize( len: Word );
  Begin
    inherited Init;
    SetFieldDefaults;
    Alloc( len );
  End; { TStrObj.InitSize }

Constructor TStrObj.InitResourceString( hInst: THandle; id: Word );
  Begin
    inherited Init;
    SetFieldDefaults;
    PutResourceString( hInst, id );
  End;

(* Load a stored object from Stream S. Our ancestor TEObject is not
   streamable and our Store method only write the string data to the
   stream. So leave the error fields at the state set by Init.
   Store does not write the terminating #0, but Alloc takes care
   of that by filling the allocated memory with #0. *)
Constructor TStrObj.Load( Var S: TStream );
  Var
    len: Integer;
  Begin
    inherited Init;
    SetFieldDefaults;
    len := 0;
    S.Read( len, Sizeof( len ));
    If ( S.Status = stOK) and ( len > 0 ) Then Begin
      Alloc( len+1 );
      If not Failed Then Begin
        S.Read( Buf[ 1 ], len );
        If S.Status = stOK Then
          SetLength( len )
        Else
          PutFill( 1, #0);
      End; { If }
    End;
  End; { TStrObj.Load }

(* Done disposes of any allocated memory and then calls the inherited Done *)
Destructor TStrObj.Done;
  Begin
    DisposeMem;
    inherited Done;
  End; { TStrObj.Done }

(* Store writes the string held in this object out to stream S. Our
   ancestor TEObject is not streamable, so there is no inherited method
   to call. Would'nt make much sense anyway, to store error codes.
   We store first an integer holding the size of the string and, if that
   is > 0, we write the characters, without the terminating #0. *)
Procedure TStrObj.Store( Var S: TStream );
  Var
    len: Word;
  Begin
    len := GetLength;
    S.Write( len, Sizeof( len ));
    If len <> 0 Then
      S.Write( Buf[ 1 ], len );
  End; { TStrObj.Store }

(* Clone provides a deep copy of this object. It returns a pointer to
   a newly allocated  object holding a copy of our own string. *)
Function TStrObj.Clone: PStrObj;
  Var
    p: PStrObj;
  Begin
    p := NewObject;
    If Assigned( p ) Then Begin
      p^.Put( Get );
      If p^.Failed Then Begin
        ForwardError( p^ );
        Dispose( p, Done );
        p := Nil;
      End; { If }
    End
    Else
      Error( errOutOfMemory, Sizeof( Self ));
    Clone := p;
  End; { TStrObj.Clone }

(* NewObject provides an empty but initialized object, freshly
   allocated on the heap. This method has to be replaced by any
   derived objects to return an instance of their own object type! *)
Function TStrObj.NewObject: PStrObj;
  Begin
    NewObject := New( PStrObj, InitBlank );
  End; { TStrObj.NewObject }

{$IFDEF DEBUG}
(* debugging helper methods *)
Var
  (* this static buffer is provided because we may not have dynamic
     memory available to build an error string in case of an error! *)
  scratchpad: Array [0..255] Of Char;

(* This replacement Error method will give an error message box for
   the three errors that may happen for a TStrObj debug version. The
   message will provide also additional info which should help to 
   identify the offending object. *)
Procedure TStrObj.Error( err, info: Word ); 
  Const
    Title: PChar= 'TStrObj Error';
    MemErr: PChar= 'Alloc: Unable to allocate %u Bytes!';
    NonSZErr: PChar='Validate: String not zero-terminated: ';
    MemOvrErr: PChar='Validate: Memory overwrite beyond end of string: ';
    BlockTooLargeErr: PChar='Fits: Requested size > 64 KByte!';
    SelfRefErr: PChar='Init/Put/Append/InsertObj: Parameter cannot be Self!';
  Var
    hWin: HWnd;
  Begin
    SetFields( err, info );
    Case err Of
      errOutOfMemory:
        wvsprintf( scratchpad, MemErr, info );
      SOERR_MemOverwrite: Begin
        StrCopy( scratchpad, MemOvrErr );
        StrLCat( scratchpad, @Buf[1], 255 );
      End;
      SOERR_NotZeroterminated: Begin
        StrCopy( scratchpad, MemOvrErr );
        StrLCat( scratchpad, @Buf[1], 255 );
      End;
      SOERR_BlockTooLarge:
        StrCopy( scratchpad, BlockTooLargeErr );
      SOERR_SelfReference:
        StrCopy( scratchpad, SelfRefErr );
      Else Begin
        wvsprintf( scratchpad,
            'TStrClass.Error: errcode= %u', err );
        wvsprintf( StrEnd( scratchpad ), ', info= %u', info );
      End;
    End; { Case }
    scratchpad[ 255 ] := #0;
    hWin := GetActiveWindow;
    MessageBox( hWin,
                scratchpad, title, MB_OK or MB_ICONSTOP  );
  End; { TStrObj.Error }

(* Validate checks if the string is properly zero-terminated and if
   the guard word is still intact. It will call Error if a suspicious
   condition is detected. *)
Procedure TStrObj.Validate;
  Var
    pGuard: ^Word;
  Begin
    If IsValid Then Begin
      pGuard := Pointer(Buf + size - 2);
      If StrEnd( Buf ) >= PChar(pGuard) Then
        Error( SOERR_NotZeroterminated, 0 )
      Else
        If pGuard^ <> SO_Guard Then
          Error( SOERR_MemOverwrite, 0 );
    End; { If }
  End; { TStrObj.Validate }

{$ENDIF}

(* field access methods, read *)
(* GetLength returns the length of the string held by this object *)
Function TStrObj.GetLength: Word;
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    If IsValid Then
      If Buf^ < #255 Then
        GetLength := Ord( Buf^ )
      Else
        GetLength := StrLen( @Buf[ 1 ] )
    Else
      GetLength := 0;
  End; { TStrObj.GetLength }

(* GetMemSize returns the amount of memory this object has allocated *)
Function TStrObj.GetMemSize: Word;
  Begin
    GetMemSize := size;
  End; { TStrObj.GetMemSize }

(* GetCapacity returns the number of characters this object can currently
   hold *)
Function TStrObj.GetCapacity: Word;
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    If IsValid Then
      GetCapacity := size - SO_Extrabytes
    Else
      GetCapacity := 0;
  End; { TStrObj.GetCapacity }

(* GetBuffer returns the buffer pointer. This should be as a read-only
   pointer! Nobody but this object should change its contents. This 
   method is provided for use by derived objects in other Units. 
   The return may be Nil!*)
Function TStrObj.GetBuffer: PChar;
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    GetBuffer := Buf;
  End; { TStrObj.GetBuffer }

(* Get returns a pointer to first character of the string held by this 
   object, or Nil, if the object is empty. This pointer should be treated
   as read-only! *)
Function TStrObj.Get: PChar; 
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    If IsValid Then
      Get := @Buf[ 1 ]
    Else
      Get := Nil;
  End; { TStrObj.Get }

(* GetString returns a _copy_ of the string currently held in this object.
   If the held string is > 255 chars long, only the first 255 will be
   returned.  *)
Function TStrObj.GetString: String;
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    If IsValid Then
      GetString := PString( Buf )^
    Else
      GetString := '';
  End; { TStrObj.GetString }

(* GetPString returns a pointer to the string currently held by this 
   object, which may be Nil. This pointer should be treated as read-only,
   never use it as target of a string operation or this objects internal
   memory organisation will be corrupted. While the compiler will think
   the target can hold 255 chars the allocated size may in fact be smaller,
   which would result in a memory overwrite. *)
Function TStrObj.GetPString: PString;
  Begin
{$IFDEF DEBUG}
    Validate;
{$ENDIF}
    GetPString := PString( Buf );
  End; { TStrObj.GetPString }

Function TStrObj.IsValid: Boolean;
  Begin
    Isvalid := Assigned( Buf );
  End; { TStrObj.IsValid }



{************************************************************
 * TStrObj.SubStr
 *
 * Parameters:
 *  aBuf: a memory buffer provided by the caller to take the substring
 *  bufsize: the size of this buffer, in bytes, including the terminating #0
 *  start: starting position of the substring to copy, counting from 0
 *  count: length of the substring to copy
 * Returns:
 *  aBuf
 * Description:
 *  Will copy a substring from the currently held string to the 
 *  buffer provided by the caller. The buffer will contain an empty
 *  string, if 
 *     - the object is empty or
 *     - start is larger than the string length.
 *  If the current string has less than count characters from start
 *  only chars up to the end of string are copied. The returned substring
 *  is always zero-terminated.
 * Error Conditions:
 *  If aBuf is Nil, nothing will be copied
 *
 *Created: 11/13/94 17:42:34 by P. Below
 ************************************************************}
Function TStrObj.SubStr( aBuf: PChar; bufsize, start, count: Word ): PChar;
  Begin
    SubStr := aBuf;
    If Assigned( aBuf ) and ( bufsize > 0 ) Then Begin
      FillChar( aBuf^, bufsize, 0 );
      If count >= bufsize  Then 
        count := bufsize - 1;
      If IsValid and ( start < GetLength ) Then
        StrLCopy( aBuf, @Buf[ start+1 ], count );
    End; { If }
  End; { TStrObj.SubStr }

{************************************************************
 * TStrObj.SubStrString
 *
 * Parameters:
 *  start: starting position of the substring to copy, counting from 1!
 *  count: length of the substring to copy
 * Returns:
 *  a copy of the substring
 * Description:
 *  Will copy a substring from the currently held string to a temp 
 *  string buffer allocated on the heap. Note that positions count
 *  from 1 here, as usual with Pascal strings!
 *  The returned string will be empty, if
 *     - the object is empty or
 *     - start is larger than the string length or
 *     - the temp string could not be allocated.
 *  If the current string has less than count characters from start
 *  only chars up to the end of string are copied. The returned substring
 *  is always limited to 255 chars, of course, but start can be > 255.
 * Error Conditions:
 *  May run out of memory on allocating the temp string.
 *
 *Created: 11/13/94 17:42:34 by P. Below
 ************************************************************}
Function TStrObj.SubStrString( start, count: Word ): String;
  Var
    pStr: ^String;
  Begin
    SubStrString := '';
    If IsValid and ( start <= GetLength ) Then Begin
      New( pStr );
      If Assigned( pStr ) Then Begin
        If count > 255 Then
          count := 255;
        If start = 0 Then
          Inc( start );
        pStr^[ 255 ] := #0;
        StrLCopy( @PStr^[ 1 ], @Buf[ start ], count );
        If pStr^[ 255 ] = #0 Then
          pStr^[ 0 ] := Chr( StrLen(@PStr^[ 1 ] ))
        Else
          pStr^[ 0 ] := #255;
        SubStrString := pStr^;
        Dispose( pStr );
      End { If }
      Else
        Error( errOutOfMemory, Sizeof( string ));
    End; { If }
  End; { TStrObj.SubStrString }

{************************************************************
 * TStrObj.SubStrObj
 *
 * Parameters:
 *  start: starting position of the substring to copy, counting from 0!
 *  count: length of the substring to copy
 * Returns:
 *  a new object containing the copy of the substring, or Nil, if
 *  it or its data buffer could not be created.
 * Description:
 *  The returned object will be empty, if
 *     - the object is empty or
 *     - start is larger than the string length or
 *     - the temp string could not be allocated.
 *  If the current string has less than count characters from start
 *  only chars up to the end of string are copied. 
 * Error Conditions:
 *  May run out of memory on allocating the new object or its string.
 *
 *Created: 11/13/94 17:42:34 by P. Below
 ************************************************************}
Function TStrObj.SubStrObj( start, count: Word ): PStrObj;
  Var
    pObj: PStrObj;
    len : Word;
  Begin
    pObj := NewObject;
    If Assigned( pObj ) Then Begin
      pObj^.Alloc( count );
      If not pObj^.Failed Then Begin
        len := StrLen( SubStr( pObj^.Get, count+1, start, count ));
        pObj^.SetLength( len );
      End { If }
      Else Begin
        ForwardError( pObj^ );
        Dispose( pObj, Done );
        pObj := Nil;
      End;
    End { If }
    Else
      Error( errOutOfMemory, Sizeof( Self ));
    SubStrObj := pObj;
  End; { TStrObj.SubStrObj }

(* methods to put data into the object *)

{************************************************************
 * TStrObj.Put
 *
 * Parameters:
 *  aStr: zero-terminated string to copy into the object
 * Description:
 *  If aStr is nil, the currently allocate storage is released,
 *  otherwise the passed string is copied to the internal storage.
 *  If necessary, the storage is expanded.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/16/94 18:01:03 by P. Below
 ************************************************************}
Procedure TStrObj.Put( aStr: PChar );
  Var
    len: TLen;
  Begin
    If Assigned( aStr ) Then Begin
      len := StrLen( aStr );
      If Fits( len, false )Then Begin
        StrCopy( Get, aStr );
        SetLength( len );
      End
    End { If }
    Else
      DisposeMem;
  End; { TStrObj.Put }

(* PutChar: copy the passed character into the object, replacing
   any previous content. Uses Put to do the work. *)
Procedure TStrObj.PutChar( c: Char );
  Var
    temp: Array [0..1] Of Char;
  Begin
    temp[ 0 ] := c;
    temp[ 1 ] := #0;
    Put( temp );
  End; { TStrObj.PutChar }

{************************************************************
 * TStrObj.PutString
 *
 * Parameters:
 *  aStr: pascal string to copy into the object
 * Description:
 *  The passed string is copied to the internal storage, replacing
 *  any previous content. 
 *  If necessary, the storage is expanded.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/16/94 18:01:03 by P. Below
 ************************************************************}
Procedure TStrObj.PutString( Const aStr: string );
  Var
    len: TLen;
  Begin
    len := Length( aStr );
    If Fits( len, false ) Then Begin
      Move( aStr, Buf^, len+1 );
      Buf[ len+1 ] := #0;
    End; 
  End; { TStrObj.PutString }

{************************************************************
 * TStrObj.PutInteger
 *
 * Parameters:
 *  aValue: an integer number to convert to a string representation
 *  fieldwidth: field width to use, same as for Str. 
 *    - if this parameter is <=0 or smaller than the number of digits
 *      needed to represent the number, it is ignored and the produced
 *      string is just long enough to represent the number.
 *    - if fieldwidth > 0, the number is right-adjusted in the field
 *      and padded with blanks to the left.
 * Description:
 *  Converts an integer number to string representation and puts it
 *  into the object, replacing any previous content.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/16/94 18:43:52 by P. Below
 ************************************************************}
Procedure TStrObj.PutInteger( aValue: LongInt; fieldwidth: Integer );
  Var
    len: Integer;
  Begin
    len := Abs( fieldwidth );
    If len < 12 Then
      len := 12;
    If Fits( len, false ) Then Begin
      PutFill( len, #0 );
      Str( aValue: fieldwidth, GetPString^ );
    End; { If }
  End; { TStrObj.PutInteger }

{************************************************************
 * TStrObj.PutReal
 *
 * Parameters:
 *  aValue: an real number to convert to a string representation
 *  fieldwidth: field width to use, same as for Str.
 *    - if this parameter is 0 or smaller than the number of digits
 *      needed to represent the number, it is ignored and the produced
 *      string is just long enough to represent the number.
 *    - if fieldwidth > 0, the number is right-adjusted in the field
 *      and padded with blanks to the left, unless exponential
 *.     format is used, in which case additional zeros are inserted
 *      before the "E" as padding to the field width (very ugly!)
 *  precision: the number of digits wanted after the decimal point.
 *    - if this value is < 0, an exponential representation is generated
 *      s1.#####Es####, where s is a sign (leading + will be suppressed)
 *      and # a digit. The number of digits after the decimal point is
 *      determined by the fieldwidth.
 *    - if this value is >=0, it generates a fixed-point representation
 *      with a maximum of 10 digits after the decimal point.
 * Description:
 *  Converts an real number to string representation and puts it
 *  into the object, replacing any previous content.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/16/94 18:43:52 by P. Below
 ************************************************************}
Procedure TStrObj.PutReal( aValue: Real; fieldwidth, precision: Integer );
  Var
    len: Integer;
  Begin
    len := Abs( fieldwidth );
    If len < 20 Then
      len := 20;
    If Fits( len, false ) Then Begin
      PutFill( len, #0 );
      Str( aValue: fieldwidth: precision, GetPString^ );
    End; { If }
  End; { TStrObj.PutReal }

(* PutObj put a copy of the string contained in aStrObj into this
   object, replacing any previous contents. *)
Procedure TStrObj.PutObj( Const aStrObj: TStrObj );
  Begin
    If @aStrObj <> @Self Then 
      Put( aStrObj.Get )
  { Else
      Error( SOERR_SELFREFERENCE, 0 )}; 
  End; { TStrObj.PutObj }

{************************************************************
 * TStrObj.PutFormat
 *
 * Parameters:
 *  fmtStr: a format string for wvsprintf
 *  data  : a record of data
 * Description:
 *  Passes the parameters unchanged to wvsprintf. Since the size
 *  of the produced string is not known in advance, we allocate
 *  a temp buffer of 1024 bytes, which is the maximum wvsprintf can
 *  handle. The formatted string is then copied into this object,
 *  replacing any previous contents.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/16/94 19:33:10 by P. Below
 ************************************************************}
Procedure TStrObj.PutFormat( fmtStr: PChar; Var data );
  Var
    tempObj: TStrObj;
  Begin
    If Assigned( fmtStr ) Then Begin
      tempObj.InitSize( 1024 );
      If not tempObj.Failed Then Begin
        wvsprintf( tempObj.Get, fmtStr, data );
        Put( tempObj.Get );
      End { If }
      Else
        ForwardError( tempObj );
      tempObj.Done;
    End; { If }
  End; { TStrObj.PutFormat }

{************************************************************
 * TStrObj.PutFill
 *
 * Parameters:
 *  count: number of repeats to produce
 *  c    : character to repeat
 * Description:
 *  Fills the object with a string consisting of count repeats of
 *  the character c.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/16/94 19:52:22 by P. Below
 ************************************************************}
Procedure TStrObj.PutFill( count: Word; c: Char );
  Begin
    If Fits( count, false ) Then Begin
      FillChar( Buf[ 1 ], count, c );
      If c = #0 Then
        SetLength( 0 )
      Else Begin
        Buf[ count+1 ] := #0;
        SetLength( count );
      End;
    End; { If }
  End; { TStrObj.PutFill }

{************************************************************
 * TStrObj.PutResourceString
 *
 * Parameters:
 *  inst: instance handle to pass to LoadString
 *  id  : string id to pass to LoadString
 * Description:
 *  Loads a string of up to 1024 chars from a resource.
 *  If the requested resource string does not exist, the object
 *  will contain an empty string (length 0).
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are nuked!
 *
 *Created: 11/16/94 20:03:54 by P. Below
 ************************************************************}
Procedure TStrObj.PutResourceString( inst: THandle; id: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitSize( 1024 );
    DeleteAll;
    If not tempObj.Failed Then Begin
      If LoadString( inst, id, tempObj.Get, 1024 ) > 0 Then
        Put( tempObj.Get )
      Else
        PutFill( 1, #0 );
    End { If }
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.PutResourceString }

{************************************************************
 * TStrObj.PutSubstr
 *
 * Parameters:
 *  aStrObj: string object to copy the substring from
 *  startAt: starting position of the substring, counting from 0
 *  count  : number of characters to copy
 * Description:
 *  Allocates the necessary space and copies the substring from
 *  the source. Differing from the other methods taking string
 *  objects aStrObj may be Self for this method!
 * Error Conditions:
 *  may run out of memory, in which case the original contents
 *  are not changed. 
 *
 *Created: 12/02/94 15:01:19 by P. Below
 ************************************************************}
Procedure TStrObj.PutSubstr( Const aStrObj: TStrObj; startAt, count: Word );
  Begin
    If @Self=@aStrObj Then Begin
      Delete( 0, startAt );
      Truncate( count );
    End 
    Else
    If aStrObj.IsValid and ( startAt < aStrObj.GetLength ) and
       Fits( count, false ) Then Begin
      aStrObj.SubStr( Get, GetCapacity+1, startAt, count );
      CalcLength;
    End; { If }
  End; { TStrObj.PutSubstr }

(* methods to change existing data.   *)
{************************************************************
 * TStrObj.Append
 *
 * Parameters:
 *  aStr: pointer to zero-terminated string to append to the
 *        contents of this object.
 * Description:
 *  Appends the passed pchar to the contents of the object,
 *  growing the allocated memory if required.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/17/94 18:24:57 by P. Below
 ************************************************************}
Procedure TStrObj.Append( aStr: PChar );
  Var
    len:  TLen;
  Begin
    If Assigned( aStr ) Then Begin
      len := StrLen( aStr );
      If len > 0 Then Begin
        len := len + GetLength;
        If Fits( len, true ) Then Begin
          StrCat( Get, aStr );
          SetLength( len );
        End; { If }
      End; { If }
    End; { If }
  End; { TStrObj.Append }

(* AppendChar appends a single char to the contents of this object.
   Uses Append to do the work. *)
Procedure TStrObj.AppendChar( c: Char );
  Var
    temp: Array [0..1] Of Char;
  Begin
    temp[ 0 ] := c;
    temp[ 1 ] := #0;
    Append( temp );
  End; { TStrObj.AppendChar }

(* AppendEOL adds a line break to the contents of the object. This method
  is virtual, so you can override it to implement a Unix-style EOL. *)
Procedure TStrObj.AppendEOL;
  Const
    CRLF: PChar= #13#10#0;
  Begin
    Append( CRLF );
  End; { TStrObj.AppendEOL }

{************************************************************
 * TStrObj.AppendString
 *
 * Parameters:
 *  aStr: the Pascal string to add to the contents of this object
 * Description:
 *  We first copy the passed string into a temp TStrObj and then
 *  call Append with the generated pchar. Of course we could do
 *  it directly, but this way is simpler (for me <g>), even if
 *  it uses more heap memory and thus runs a slightly higher risk
 *  of running out of memory. The target string is not limited to 
 *  255 characters, unlike in a normal Pascal string operation!
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/17/94 18:35:27 by P. Below
 ************************************************************}
Procedure TStrObj.AppendString( Const aStr: String );
  Var
    tempObj: TStrObj;
  Begin
    If Length( aStr ) > 0 Then Begin
      tempObj.InitString( aStr );
      If not tempObj.Failed Then 
        Append( tempObj.Get )
      Else
        ForwardError( tempObj );
      tempObj.Done;
    End; { If }
  End; { TStrObj.AppendString }

(* Append the string representation of the passed number to this
   objects contents. Uses the same temp object trick as AppendString. 
   See PutInteger for a description of the parameters. *)
Procedure TStrObj.AppendInteger( aValue: LongInt; fieldwidth: Integer );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutInteger( aValue, fieldwidth );
    If not tempObj.Failed Then 
      Append( tempObj.Get  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.AppendInteger }

(* Append the string representation of the passed number to this
   objects contents. Uses the same temp object trick as AppendString. 
   See PutReal for a description of the parameters. *)
Procedure TStrObj.AppendReal( aValue: Real; fieldwidth, precision: Integer );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutReal( aValue, fieldwidth, precision );
    If not tempObj.Failed Then 
      Append( tempObj.Get  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.AppendReal }

(*  pretty self-evident what this one does, is'nt it? *)
Procedure TStrObj.AppendObj( Const aStrObj: TStrObj );
  Begin
    If @aStrObj <> @Self Then 
      Append( aStrObj.Get )
    Else
      Error( SOERR_SELFREFERENCE, 0 );
  End; { TStrObj.AppendObj }

(* Append a formatted string ( wvsprintf-style) to this
   objects contents. Uses the same temp object trick as AppendString. 
   See PutFormat for a description of the parameters. *)
Procedure TStrObj.AppendFormat( fmtStr: PChar; Var data );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutFormat( fmtStr, data );
    If not tempObj.Failed Then 
      Append( tempObj.Get  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.AppendFormat }

(* Append a string of count characters c to this
   objects contents. Uses the same temp object trick as AppendString. 
   See PutFill for a description of the parameters. *)
Procedure TStrObj.AppendFill( count: Word; c: Char );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutFill( count, c );
    If not tempObj.Failed Then 
      Append( tempObj.Get  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.AppendFill }

{************************************************************
 * TStrObj.AppendPad
 *
 * Parameters:
 *  withC: char to use for padding
 *  toLength: length to pad the current string to
 * Description:
 *  This is a wrapper for a special case for AppendFill. It will
 *  extend the current string with the passed char until it has 
 *  a length of toLength characters. 
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/17/94 23:15:19 by P. Below
 ************************************************************}
Procedure TStrObj.AppendPad( withC: Char; toLength: Word );
  Var
    len: Word;
  Begin
    len := GetLength;
    If toLength > len Then
      AppendFill( toLength-len, withC );
  End; { TStrObj.AppendPad }


(* Append the requested resource string to this
   objects contents. Uses the same temp object trick as AppendString. 
   See PutResourceString for a description of the parameters. *)
Procedure TStrObj.AppendResourceString( inst: THandle; id: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutResourceString( inst, id );
    If not tempObj.Failed Then 
      Append( tempObj.Get  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.AppendResourceString }

{************************************************************
 * TStrObj.Insert
 *
 * Parameters:
 *  aStr : pointer to a zero-terminated string to insert
 *  atPos: position to insert the string at. Counts from 0.
 * Description:
 *  Inserts the passed string into the string held by this object.
 *  All characters after and including the one at position atPos
 *  (which corresponds to Buf[ atPos+1 ] in our internal representation)
 *   are moved to higher addresses to make room for the new text.
 *   Memory grows as needed to fit the combined string.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/17/94 19:22:47 by P. Below
 ************************************************************}
Procedure TStrObj.Insert( aStr: Pchar; atPos: Word );
  Var
    len: TLen;
    thisLen, insLen: Word;
  Begin
    If Assigned( aStr ) Then Begin
      insLen := StrLen( aStr );
      If insLen > 0 Then Begin
        thisLen := GetLength;
        If atPos >= thisLen Then
          (* requested position beyond end of current contents, so
             just append aStr *)
          Append( aStr )
        Else Begin
          len := TLen(insLen) + TLen( thisLen );
          (* cast to prevent overflow in debugging version *)
          If Fits( len, true ) Then Begin
            (* move the substring after the insertion point, including
               the terminating #0 for insLen positions up.
               Rememeber: Buf[ 1 ] is the start of our internal pchar but
               atPos counts from 0. *)
            StrMove( @Buf[atPos+insLen+1], @Buf[ atPos+1 ], thisLen+1-atPos );
            (* copy the string to insert into the gap we just made,
               without its terminating #0. *)
            StrMove( @Buf[ atPos+1 ], aStr, insLen );
            SetLength( len );
          End; { If }
        End; { Else }
      End; { If }
    End; { If }
  End; { TStrObj.Insert }

{************************************************************
 * TStrObj.InsertChar
 *
 * Parameters:
 *  c : character to insert
 *  atPos: position to insert the character at. Counts from 0.
 * Description:
 *  Build a temp string and then calls Insert to do the work.
 *  All characters after and including the one at position atPos
 *  (which corresponds to Buf[ atPos+1 ] in our internal representation)
 *  are moved one position to higher addresses to make room for the new 
 *  character.
 *  Memory grows as needed to fit the expanded string.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/19/94 14:56:05 by P. Below
 ************************************************************}
Procedure TStrObj.InsertChar( c: char; atPos: Word );
  Var
    temp: Array [0..1] Of Char;
  Begin
    temp[ 0 ] := c;
    temp[ 1 ] := #0;
    Insert( temp, atPos );
  End; { TStrObj.InsertChar }

{************************************************************
 * TStrObj.InsertString
 *
 * Parameters:
 *  aStr : a Pascal string to insert
 *  atPos: position to insert the string at. Counts from 1 (!!).
 * Description:
 *  Inserts the passed string into the string held by this object.
 *  All characters after and including the one at position atPos
 *  (which corresponds to Buf[ atPos+1 ] in our internal representation)
 *  are moved to higher addresses to make room for the new text.
 *  Memory grows as needed to fit the combined string.
 *  Takes the wimpy way out by first putting the passed string into 
 *  a temp object and the calling Insert to do the work.
 *  Unlike the build-in string functions atPos can be > 255 here and
 *  the target string is also not limited to 255 chars.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/17/94 19:22:47 by P. Below
 ************************************************************}
Procedure TStrObj.InsertString( Const aStr: String; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    If Length( aStr ) > 0 Then Begin
      If atPos > 0 Then 
        Dec( atPos );
      tempObj.InitString( aStr );
      If not tempObj.Failed Then 
        Insert( tempObj.Get , atPos )
      Else
        ForwardError( tempObj );
      tempObj.Done;
    End; { If }
  End; { TStrObj.InsertString }

{************************************************************
 * TStrObj.InsertInteger
 *
 * Parameters:
 *  aValue: an integer number to convert to a string representation
 *  fieldwidth: field width to use, same as for Str. 
 *    - if this parameter is <=0 or smaller than the number of digits
 *      needed to represent the number, it is ignored and the produced
 *      string is just long enough to represent the number.
 *    - if fieldwidth > 0, the number is right-adjusted in the field
 *      and padded with blanks to the left.
 *  atPos: position to insert the number at. Counts from 0.
 * Description:
 *  Converts an integer number to string representation and inserts it
 *  into the object at the requested position, moving any previous content
 *  after that position up to higher addresses. Storage space will grow
 *  as needed.
 *  Takes the wimpy way out by first putting the passed number into 
 *  a temp object and the calling Insert to do the work.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/19/94 15:20:03 by P. Below
 ************************************************************}
Procedure TStrObj.InsertInteger( aValue: LongInt; fieldwidth: Integer ; atPos: Word);
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutInteger( aValue, fieldwidth );
    If not tempObj.Failed Then 
      Insert( tempObj.Get , atPos )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.InsertInteger }

{************************************************************
 * TStrObj.InsertReal
 *
 * Parameters:
 *  aValue: an integer number to convert to a string representation
 *  fieldwidth: field width to use, same as for Str. 
 *    - if this parameter is <=0 or smaller than the number of digits
 *      needed to represent the number, it is ignored and the produced
 *      string is just long enough to represent the number.
 *    - if fieldwidth > 0, the number is right-adjusted in the field
 *      and padded with blanks to the left.
 *  precision: the number of digits wanted after the decimal point.
 *    - if this value is < 0, an exponential representation is generated
 *      s1.#####Es####, where s is a sign (leading + will be suppressed)
 *      and # a digit. The number of digits after the decimal point is
 *      determined by the fieldwidth.
 *    - if this value is >=0, it generates a fixed-point representation
 *      with a maximum of 10 digits after the decimal point.
 *  atPos: position to insert the number at. Counts from 0.
 * Description:
 *  Converts a real number to string representation and inserts it
 *  into the object at the requested position, moving any previous content
 *  after that position up to higher addresses. Storage space will grow
 *  as needed.
 *  Takes the wimpy way out by first putting the passed number into 
 *  a temp object and the calling Insert to do the work.
 * Error Conditions:
 *  May run out of memory if the storage space has to grow. In
 *  this case the original contents of the object are retained.
 *
 *Created: 11/19/94 15:20:03 by P. Below
 ************************************************************}
Procedure TStrObj.InsertReal( aValue: Real; fieldwidth, precision: Integer ; atPos: Word);
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutReal( aValue, fieldwidth, precision );
    If not tempObj.Failed Then 
      Insert( tempObj.Get , atPos )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.InsertReal }

(* Insert contents of passed object into self at requested position;
   uses Insert to do all the work. atPos counts from 0! *)
Procedure TStrObj.InsertObj( Const aStrObj: TStrObj; atPos: Word );
  Begin
    If @aStrObj <> @Self Then 
      Insert(  aStrObj.Get, atPos )
    Else
      Error( SOERR_SELFREFERENCE, 0 );
  End; { TStrObj.InsertObj }

{************************************************************
 * TStrObj.InsertFormat
 *
 * Parameters:
 *  fmtStr: a format string for wvsprintf
 *  data  : a record of data
 *  atPos: position to insert the number at. Counts from 0.
 * Description:
 *  Formats the passed data according to the fmtStr and inserts the
 *  resulting string into self at the requested position. Uses a
 *  temporary object to make life simpler. 
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/16/94 19:33:10 by P. Below
 ************************************************************}
Procedure TStrObj.InsertFormat( fmtStr: PChar; Var data; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutFormat( fmtStr, data );
    If not tempObj.Failed Then 
      Insert( tempObj.Get , atPos  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.InsertFormat }

{************************************************************
 * TStrObj.InsertFill
 *
 * Parameters:
 *  count: number of repeats to produce
 *  c    : character to repeat
 *  atPos: position to insert the number at. Counts from 0.
 * Description:
 *  Inserts a string consisting of count repeats of the character c
 *  at the requested position, moving all subsequent characters up.
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/19/94 15:34:00 by P. Below
 ************************************************************}
Procedure TStrObj.InsertFill( count: Word; c: Char; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutFill( count, c );
    If not tempObj.Failed Then 
      Insert( tempObj.Get , atPos  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.InsertFill }

{************************************************************
 * TStrObj.InsertResourceString
 *
 * Parameters:
 *  inst: instance handle to pass to LoadString
 *  id  : string id to pass to LoadString
 *  atPos: position to insert the number at. Counts from 0.
 * Description:
 *  Loads a string of up to 1024 chars from a resource and inserts
 *  it into self at the requested position. 
 * Error Conditions:
 *  May run out of memory, in which case the original contents of
 *  this object are not changed.
 *
 *Created: 11/16/94 20:03:54 by P. Below
 ************************************************************}
Procedure TStrObj.InsertResourceString( inst: THandle; id: Word; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutResourceString( inst, id );
    If not tempObj.Failed Then 
      Insert( tempObj.Get , atPos  )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.InsertResourceString }

{************************************************************
 * TStrObj.Delete
 *
 * Parameters:
 *  atPos: position to start deleting. Counts from 0.
 *  count: number of characters to delete
 * Description:
 *  Deletes count characters starting at position atPos in the
 *  currently held string.
 *  If atPos is beyond the end of string, nothing will be deleted.
 *  If less than count characters are present after atPos, all
 *  characters from atPos to the end of string will be deleted.
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 15:38:01 by P. Below
 ************************************************************}
Procedure TStrObj.Delete( atPos, count: Word ); 
  Var
    len: Word;
  Begin
    If IsValid and (count > 0) Then Begin
      len := GetLength;
      If atPos >= len Then 
        Exit; (* start position is beyond end of string, do nothing *)

      If count >= ( len-atPos ) Then Begin
        (* delete all of the string beyond and including atPos *)
        Buf[ atPos+1 ] := #0;
        SetLength( atPos );
      End { If }
      Else Begin
        StrMove( @Buf[ atPos+1 ], @Buf[ atPos+1+count ],
                 len - count -atPos +1 );
        SetLength( len-count );
      End; { If }
    End; { If }
  End; { TStrObj.Delete }

{************************************************************
 * TStrObj.DeleteAll
 *
 * Parameters:
 *  none
 * Description:
 *  Delete the current string by overwriting it with #0
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 15:44:51 by P. Below
 ************************************************************}
Procedure TStrObj.DeleteAll;
  Begin
    PutFill( GetCapacity, #0 );
  End; { TStrObj.DeleteAll }

{************************************************************
 * TStrObj.Truncate
 *
 * Parameters:
 *  atPos: position to truncate the string at, counting from 0
 * Description:
 *  replaces the character at position atPos with a #0, thus 
 *  truncating the string to atPos characters
 * Error Conditions:
 *  none
 *
 *Created: 11/28/94 18:08:19 by P. Below
 ************************************************************}
Procedure TStrObj.Truncate( atPos: Word );
  Begin
    If IsValid Then 
      If atPos < GetLength Then Begin
        Buf[ atPos+1 ] := #0;
        SetLength( atPos );
      End; { If }
  End; { TStrObj.Truncate }

{************************************************************
 * TStrObj.RTrim
 *
 * Parameters:
 *  none
 * Description:
 *  removes any whitespace from the end of the string. We count
 *  anything with ascii codes <= 32 as whitespace, excluding #0.
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 18:19:23 by P. Below
 ************************************************************}
Procedure TStrObj.RTrim;
  Var
    n: Word;
    p: PChar;
  Begin
    If IsValid Then Begin
      p := Get;
      n := GetLength;
      While ( n>0 ) and ( p[ n-1 ] <= ' ' ) Do 
        Dec( n );
      If n < GetLength Then 
        Truncate( n );
    End; { If }
  End; { TStrObj.RTrim }

{************************************************************
 * TStrObj.LTrim
 *
 * Parameters:
 *  none
 * Description:
 *  removes any whitespace from the start of the string. We count
 *  anything with ascii codes <= 32 as whitespace, excluding #0.
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 18:19:23 by P. Below
 ************************************************************}
Procedure TStrObj.LTrim;
  Var
    n: Word;
    p: PChar;
  Begin
    If IsValid Then Begin
      p := Get;
      n := 0;
      While ( p[ n ] > #0 ) and ( p[ n ] <= ' ' ) Do
        Inc( n );
      If n > 0 Then
        Delete( 0, n );
    End;
  End; { TStrObj.LTrim }

{************************************************************
 * TStrObj.Trim
 *
 * Parameters:
 *  none
 * Description:
 *  removes any whitespace from the start and end of the string. We count
 *  anything with ascii codes <= 32 as whitespace, excluding #0.
 * Error Conditions:
 *  none
 *
 *Created: 11/30/94 18:19:23 by P. Below
 ************************************************************}
Procedure TStrObj.Trim;
  Begin
    LTrim;
    RTrim
  End; { TStrObj.Trim }

{************************************************************
 * TStrObj.Compare
 *
 * Parameters:
 *  aStr: pointer to a zero-terminated string to compare with self
 *  caseSensitive: flag stating wether the comparison should be case-
 *        sensitive or not.
 * Returns:
 *   < 0, if the held string is lexically smaller than aStr,
 *   = 0, if it is equal,
 *   > 0, if it is larger.
 * Description:
 *  The comparison will Windows API functions, that support national
 *  language characters, if WINDOWS is defined. Otherwise, it will use
 *  the functions supplied by the Strings Unit.
 *  A Nil pointer will be treated as "smaller" than any string, even an
 *  empty one. If aStr _and_ out buffer are both Nil, they will be 
 *  considered equal!
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 15:46:06 by P. Below
 ************************************************************}
Function TStrObj.Compare( aStr: PChar; caseSensitive: Boolean ): Integer;
  Begin
    If IsValid Then 
      If Assigned( aStr ) Then 
        If caseSensitive Then
{$IFDEF WINDOWS}
          Compare := lstrcmpi( Get, aStr )
{$ELSE}
          Compare := StrIComp( Get, aStr )
{$ENDIF}
        Else
{$IFDEF WINDOWS}
          Compare := lstrcmp( Get, aStr )
{$ELSE}
          Compare := StrComp( Get, aStr )
{$ENDIF}
      Else
        Compare := 1
    Else
      If Assigned( aStr ) Then
        Compare := -1
      Else
        Compare := 0;
  End; { TStrObj.Compare }

{************************************************************
 * TStrObj.CompareString
 *
 * Parameters:
 *  aStr: a Pascal string to compare with self
 *  caseSensitive: flag stating wether the comparison should be case-
 *        sensitive or not.
 * Returns:
 *   < 0, if the held string is lexically smaller than aStr,
 *   = 0, if it is equal,
 *   > 0, if it is larger.
 * Description:
 *  Like Compare, this function compares the passed parameter to the
 *  currently held string. There is a small difference to a direct
 *  comparison of two pascal strings ( if str1 = str2 then... ) as done
 *  by the compiler: if the current string is longer than 255 characters
 *  and the passed string is exactly 255 chars long, this function will
 *  not judge them equal, even if the first 255 chars are the same!
 * Error Conditions:
 *  May run out of memory for the temp TStrObj that is used to convert
 *  the passed pascal string to a Pchar for Compare. In this case it
 *  will report an error and return High( Integer ).
 *
 *Created: 11/19/94 16:02:23 by P. Below
 ************************************************************}
Function TStrObj.CompareString( Const aStr: String; 
                                caseSensitive: Boolean): Integer;
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitString( aStr );
    If not tempObj.Failed Then
      CompareString := Compare( tempObj.Get , caseSensitive )
    Else Begin
      ForwardError( tempObj );
      CompareString := High( Integer );
    End;
    tempObj.Done;
  End; { TStrObj.CompareString }

{************************************************************
 * TStrObj.CompareString
 *
 * Parameters:
 *  aStrObj: an object to compare with self
 *  caseSensitive: flag stating wether the comparison should be case-
 *        sensitive or not.
 * Returns:
 *   < 0, if the held string is lexically smaller than aStr,
 *   = 0, if it is equal,
 *   > 0, if it is larger.
 * Description:
 *  Uses Compare to compare the passed objects string to out own.
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 16:02:23 by P. Below
 ************************************************************}
Function TStrObj.CompareObj( Const aStrObj: TStrObj; 
                             caseSensitive: Boolean): Integer;
  Begin
    CompareObj := Compare( aStrObj.Get, caseSensitive );
  End; { TStrObj.CompareObj }

{************************************************************
 * TStrObj.Replace
 *
 * Parameters:
 *  aPattern: pointer to a zero-terminated string to replace in self.
 *            If this parameter is Nil or the string is empty, nothing
 *            will be done.
 *  aStr    : pointer to a zero-terminated string to replace it with.
 *            If this parameter is Nil or the string is empty, the
 *            pattern string will be deleted from self.
 * Returns:
 *   true, if a replacement was actualy made, false otherwise
 * Description:
 *  Searches for the passed pattern string (case-sensitive!), and, if
 *  it is found in our current string, deletes it and inserts aStr
 *  at its position.
 *  This function could be implemented by calling Delete and Insert. This,
 *  however, requires three StrMoves in total, where two will do with a
 *  little more preparation. Since we need to do most of the required
 *  calculations anyway to make sure the replacement will fit before we
 *  delete the pattern string, we do not use Delete/Insert here.
 * Error Conditions:
 *  May run out of memory, in which case the the original contents of this
 *  obj are not changed.
 *
 *Created: 11/19/94 16:27:01 by P. Below
 ************************************************************}
Function TStrObj.Replace( aPattern, aStr: PChar ): Boolean;
  Var
    patternLen, aStrLen, foundLoc, charsToMove: Word;
    len: TLen;
    p  : PChar;
  Begin
    Replace := True;
    If Assigned( aPattern ) Then Begin
      patternLen := StrLen( aPattern );
      If patternLen > 0 Then Begin
        foundLoc := Pos( aPattern );
        If foundLoc <> NOT_FOUND Then Begin
          (* we found the pattern, now check the replacement string *)
          If Assigned( aStr ) Then
            aStrLen := StrLen( aStr )
          Else
            aStrLen := 0;
          If aStrLen = 0 Then
            (* no problem, just delete the pattern if there's nothing
               to replace it. *)
            Delete( patternLen, foundLoc )
          Else Begin
            (* check if we need to grow, if yes, try to do it *)
            len := TLen( GetLength ) + aStrLen  - patternLen ;
            If aStrLen > patternLen Then Begin
              If not Fits( len, true ) Then
                Exit;  (* grmbl! out of memory! *)
            End;
            (* ok, replacement will fit *)
            p := At( foundLoc );
            If patternLen <> aStrLen Then Begin
              (* move string after pattern, including the terminating #0 *)
              charsToMove := GetLength - foundLoc - patternLen + 1;
              StrMove( p+aStrLen, p+patternlen, charsToMove );
            End;
            StrMove( p, aStr, aStrLen );
            CalcLength;
          End; { Else }
        End { If }
        Else
          Replace := False;
      End; { If }
    End; { If }
  End; { TStrObj.Replace }

{************************************************************
 * TStrObj.ReplaceString
 *
 * Parameters:
 *  aPattern: Pascal string to replace in self.
 *            If this string is empty, nothing will be done.
 *  aStr    : Pascal string to replace it with.
 *            If this string is empty, the pattern string will be
 *            deleted from self.
 * Returns:
 *   true, if a replacement was actualy made, false otherwise
 * Description:
 *  Searches for the passed pattern string (case-sensitive!), and, if
 *  it is found in our current string, deletes it and inserts aStr
 *  at its position.
 *  Converts the two PAscal strings to pchars via two temp objects and
 *  calls Replace to do the work.
 *  Unlike the build-in string functions the pattern be found at
 *  positions > 255 here and
 *  the target string is also not limited to 255 chars.
 * Error Conditions:
 *  May run out of memory, in which case the the original contents of this
 *  obj are not changed.
 *
 *Created: 11/19/94 16:27:01 by P. Below
 ************************************************************}
Function TStrObj.ReplaceString( Const aPattern, aStr: String ): Boolean;
  Var
    temp1, temp2: TStrObj;
  Begin
    ReplaceString := False;
    If Length( aPattern ) > 0 Then Begin
      temp1.InitString( aPattern );
      If temp1.Failed Then
        ForwardError( temp1 )
      Else Begin
        temp2.InitString( aStr );
        If temp2.Failed Then
          ForwardError( temp2 )
        Else
          ReplaceString := Replace( temp1.Get, temp2.Get );
      End;
      temp2.Done;
      temp1.Done;
    End; { If }
  End; { TStrObj.ReplaceString }


{************************************************************
 * TStrObj.Overwrite
 *
 * Parameters:
 *  aStr: pointer to a zero-terminated string. If Nil, nothing
 *        will be done.
 *  atPos: position in our string to start at, counts from 0.
 * Description:
 *  Will overwrite the current string with the passed string, 
 *  starting at position atPos. If this position is beyond the end
 *  of the current string, aStr is just appended. If the overwrite
 *  goes beyond the end of string, the current string is expanded.
 * Error Conditions:
 *  May run out of memory, if the storage needs to grow. In this
 *  case the the original contents of this obj are not changed.
 *
 *Created: 11/19/94 17:55:52 by P. Below
 ************************************************************}
Procedure TStrObj.Overwrite( aStr: PChar; atPos: Word );
  Var
    len: TLen;
    thisLen, aStrLen: Word;
  Begin
    If Assigned( aStr ) Then Begin
      aStrLen := StrLen( aStr );
      If aStrLen > 0 Then Begin
        thisLen := GetLength;
        If atPos >= thisLen Then
          Append( aStr )
        Else Begin
          len := atPos + aStrLen;
          If Fits( len, true ) Then Begin
            If len > thisLen Then
              (* need to copy termiating #0, too *)
              Inc( aStrLen );
            StrMove( At( atPos ), aStr, aStrLen );
            If len > thisLen Then
              (* need to set new length  *)
              SetLength( len );
          End; { If }
        End; { Else }
      End; { If }
    End; { If }
  End; { TStrObj.Overwrite }

{************************************************************
 * TStrObj.OverwriteChar
 *
 * Parameters:
 *  c: char to put into the current string
 *  atPos: position to put it at, counts from 0
 * Description:
 *  If the object contains a string and if atPos is inside this
 *  string, the passed character will replace the current one
 *  in our string. If c is #0, the string length will be corrected
 *  accordingly. If atPos is beyound the end of string we append c,
 *  but atPos may be =GetLength, in which case we overwrite the 
 *  terminating #0 and recalculate the string length! This may come
 *  in handy in situations in which one needs to chop off a part of a 
 *  string and restore it later. See TFilename.CreatePath for an example.
 *  Note that we do not allow the safeguard #0 ( atPOs = GetCapacity ) to
 *  be overwritten, so the string will stay zero-terminated!
 * Error Conditions:
 *  May run out of memory if we need to append c.
 *
 *Created: 11/20/94 13:28:31 by P. Below
 ************************************************************}
Procedure TStrObj.OverwriteChar( c: char; atPos: Word );
  Var
    len: Word;
  Begin
    If IsValid Then Begin
      len := GetLength;
      If ( atPos < len ) Then Begin
        Buf[ atPos+1 ] := c;
        If c = #0 Then
          SetLength( atPos );
      End { If }
      Else If atPos = len Then Begin
        { this is a special case, it overwrites the terminating #0
          of the current string. We allow this because it comes in handy
          in certain situations. To make doubly sure the resulting 
          string is properly zero-terminated we set the last byte in the
          buffer to #0. }
        If c <> #0 Then Begin
          Buf[ atPos+1 ] := c;
          Buf[ GetCapacity+1 ] := #0;
          CalcLength;
        End;
      End { If }
      Else
        AppendChar( c );
    End;
  End; { TStrObj.OverwriteChar }

{************************************************************
 * TStrObj.OverwriteString
 *
 * Parameters:
 *  aStr: a Pascal string. If Nil, nothing
 *        will be done.
 *  atPos: position in our string to start at, counts from 1.
 * Description:
 *  Will overwrite the current string with the passed string, 
 *  starting at position atPos. If this position is beyond the end
 *  of the current string, aStr is just appended. If the overwrite
 *  goes beyond the end of string, the current string is expanded.
 *  atPos may be > 255 and the resulting string is also not limited
 *  to 255 chars!
 * Error Conditions:
 *  May run out of memory, in this
 *  case the the original contents of this obj are not changed.
 *
 *Created: 11/19/94 17:55:52 by P. Below
 ************************************************************}
Procedure TStrObj.OverwriteString( Const aStr: String; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitString( aStr );
    If not tempObj.Failed Then Begin
      If atPos > 0 Then Dec( atPos );
      Overwrite( tempObj.Get , atPos );
    End
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.OverwriteString }

{************************************************************
 * TStrObj.OverwriteFill
 *
 * Parameters:
 *  count: number of characters to write
 *  c    : character to use
 *  atPos: startig position in out string
 * Description:
 *  Will overwrite the current string with count characters, c,
 *  starting at atPos. If atPos is beyond the end of string, the
 *  characters are appended. If the write overruns the old end of
 *  string, the string is expanded.
 * Error Conditions:
 *  May run out of memory, in this
 *  case the the original contents of this obj are not changed.
 *
 *Created: 11/19/94 18:10:21 by P. Below
 ************************************************************}
Procedure TStrObj.OverwriteFill( count: Word; c: Char; atPos: Word );
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitBlank;
    tempObj.PutFill(  count, c );
    If not tempObj.Failed Then
      Overwrite( tempObj.Get , atPos )
    Else 
      ForwardError( tempObj );
    tempObj.Done;
  End; { TStrObj.OverwriteFill }

  (* no object version provided here *)

(* conversion methods *)
{************************************************************
 * TStrObj.ToUpper
 *
 * Parameters:
 *  none
 * Description:
 *  Converts the current string to uppercase, using either the
 *  Windows API function AnsiUpper or StrUpper.
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 18:23:06 by P. Below
 ************************************************************}
Procedure TStrObj.ToUpper;
  Begin
    If IsValid Then
{$IFDEF WINDOWS }
      AnsiUpper( Get );
{$ELSE}
      StrUpper( Get );
{$ENDIF}
  End; { TStrObj.ToUpper }

{************************************************************
 * TStrObj.ToLower
 *
 * Parameters:
 *  none
 * Description:
 *  Converts the current string to lowercase, using either the
 *  Windows API function AnsiUpper or StrUpper.
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 18:23:06 by P. Below
 ************************************************************}
Procedure TStrObj.ToLower;
  Begin
    If IsValid Then
{$IFDEF WINDOWS }
      AnsiLower( Get );
{$ELSE}
      StrLower( Get );
{$ENDIF}
  End; { TStrObj.ToLower }

{************************************************************
 * TStrObj.ToOEM
 *
 * Parameters:
 *  none
 * Description:
 *  Converts the string from ANSI to OEM character set under
 *  Windows. Is empty if not compiled for Windows
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 18:33:39 by P. Below
 ************************************************************}
Procedure TStrObj.ToOEM;
  Begin
{$IFDEF WINDOWS }
    If IsValid Then
      AnsiToOem( Get, Get );
{$ENDIF}
  End; { TStrObj.ToOEM }

{************************************************************
 * TStrObj.ToOEM
 *
 * Parameters:
 *  none
 * Description:
 *  Converts the string from OEM to ANSI character set under
 *  Windows. Is empty if not compiled for Windows
 * Error Conditions:
 *  none
 *
 *Created: 11/19/94 18:33:39 by P. Below
 ************************************************************}
Procedure TStrObj.ToANSI;
  Begin
{$IFDEF WINDOWS }
    If IsValid Then
      OemToAnsi( Get, Get );
{$ENDIF}
  End; { TStrObj.ToANSI }

(* the following two methods report errors via the Failed method
   inherited from TEObject. A conversion error will NOT trigger
   the Error method! *)

{************************************************************
 * TStrObj.ToInteger
 *
 * Parameters:
 *  none
 * Returns:
 *  an integer number read from the internal string or 0, in case
 *  of error
 * Description:
 *  Isolates the first sequence of digits it finds in our string and
 *  converts it to a number.
 *  This is a bit tricky because Pascals Var function is quite un-
 *  forgiving if anything other than a digit is found after the number,
 *  but it will accept whitespace in front. If Val failes we try to 
 *  replace the offending character with #0 and try a second time. But
 *  we also require that the string has nothing but whitespace before the
 *  first digit, except a sign!
 * Error Conditions:
 *  will fail if the string does not contain a number matching the require-
 *  ments. This is a non-fatal error, so Error will not be called!
 *
 *Created: 11/19/94 18:43:46 by P. Below
 ************************************************************}
Function  TStrObj.ToInteger: LongInt;
  Var
    temp: LongInt;
    err : Integer;
    ch  : Char;
    p   : Pchar;
  Begin
    temp := 0;
    If IsValid and ( GetLength > 0 ) Then Begin
      Val( Get, temp, err );
      If err <> 0 Then Begin
        (* blast! failed simple conversion, try to put #0 on offending
           character. err returns index counting from 1 *)
        Dec( err );
        p := At( err );
        ch := p^;
        p^:= #0;
        Val( Get, temp, err );
        p^:= ch;
        If err <> 0 Then Begin
          SetFields( SOERR_ConversionFailed, err );
          temp := 0
        End;
      End;
    End { If }
    Else
      SetFields( SOERR_ConversionFailed, 0 );
    ToInteger := temp;
  End; { TStrObj.ToInteger }

{************************************************************
 * TStrObj.ToReal
 *
 * Parameters:
 *  none
 * Returns:
 *  an real number read from the internal string or 0.0, in case
 *  of error
 * Description:
 *  Isolates the first sequence of digits it finds in our string and
 *  converts it to a number.
 *  This is a bit tricky because Pascals Var function is quite un-
 *  forgiving if anything other than a digit is found after the number,
 *  but it will accept whitespace in front. If Val failes we try to 
 *  replace the offending character with #0 and try a second time. But
 *  we also require that the string has nothing but whitespace before the
 *  first digit, except a sign!
 * Error Conditions:
 *  will fail if the string does not contain a number matching the require-
 *  ments. This is a non-fatal error, so Error will not be called!
 *
 *Created: 11/19/94 18:43:46 by P. Below
 ************************************************************}
Function  TStrObj.ToReal: Real;
  Var
    temp: Real;
    err : Integer;
    ch  : Char;
    p   : Pchar;
  Begin
    temp := 0.0;
    If IsValid and ( GetLength > 0 ) Then Begin
      Val( Get, temp, err );
      If err <> 0 Then Begin
        (* blast! failed simple conversion, try to put #0 on offending
           character. err returns index counting from 1 *)
        Dec( err );
        p := At( err );
        ch := p^;
        p^:= #0;
        Val( Get, temp, err );
        p^:= ch;
        If err <> 0 Then Begin
          SetFields( SOERR_ConversionFailed, err );
          temp := 0.0;
        End;
      End;
    End { If }
    Else 
      SetFields( SOERR_ConversionFailed, 0 );
    ToReal := temp;
  End; { TStrObj.ToReal }

(* methods to search the data *)

{************************************************************
 * TStrObj.Match
 *
 * Parameters:
 *  anExpression: search expression to match against. This
 *                expression may contain * and ? as wildcards,
 *                ? matches any character, * any sequence of 
 *                characters.
 * Returns:
 *  True if the current string matches the pattern, false if it
 *  does not or if the current object is empty or the parameter is
 *  Nil.
 * Description:
 *  Uses a wildcard match function filched from the TopSpeed 
 *  Modula-2 runtime library and rewritten for Pascal.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:01:55 by P. Below
 ************************************************************}
Function TStrObj.Match( anExpression: PChar ): Boolean;
  Function Rmatch(s:Pchar;i:Integer;p:PChar;j:Integer):Boolean;
  Var
    matched        : Boolean;
    k              : Integer;
  Begin
    If p[0]=#0 Then
      Rmatch :=  TRUE
    Else
      While TRUE Do
        If (s[i]=#0) and (p[j]=#0) Then Begin
          Rmatch :=  TRUE;
          Exit
        End
        Else
          If p[j]=#0 Then Begin
            Rmatch :=  FALSE;
            Exit
          End
          Else
            If (p[j]='*') Then Begin
              k :=i;
              If (p[j+1]=#0) Then Begin
                Rmatch :=  TRUE;
                Exit
              End
              Else
                While TRUE Do Begin
                  matched := Rmatch(s,k,p,j+1);
                  If matched OR (s[k]=#0) Then Begin
                    Rmatch :=  matched;
                    Exit;
                  End; (*If*)
                  INC(k);
                End; (* While*)
            End (*If*)
            Else
              If ((p[j]='?') and (s[i]<>#0)) OR
                 (UpCase(p[j])=UpCase(s[i])) Then Begin
                INC(i);
                INC(j);
              End
              Else Begin
                Rmatch :=  FALSE;
                Exit
              End;
  End; { RMatch }

  Begin (*Match*)
    If IsValid and Assigned( anExpression ) Then
      Match :=  Rmatch(Get,0,anExpression,0)
    Else
      Match := False
  End; { TStrObj.Match }

{************************************************************
 * TStrObj.Match
 *
 * Parameters:
 *  anExpression: search expression to match against. This
 *                expression may contain * and ? as wildcards,
 *                ? matches any character, * any sequence of 
 *                characters.
 * Returns:
 *  True if the current string matches the pattern, false if it
 *  does not or if the current object is empty or the parameter is
 *  Nil or an error occured.
 * Description:
 *  Uses a temp tStrObj to convert the parameter to pchar and then 
 *  calls Match to do the work. The mach is not limited to the first
 *  255 characters of our string!
 * Error Conditions:
 *  May run out of memory for the temp object, i which case False
 *  is returned.
 *
 *Created: 11/20/94 13:01:55 by P. Below
 ************************************************************}
Function TStrObj.MatchString( Const anExpression: String ): Boolean;
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitString( anExpression );
    If not tempObj.Failed Then
      MatchString := Match( tempObj.Get )
    Else Begin
      ForwardError( tempObj );
      MatchString := False;
    End;
    tempObj.Done;
  End; { TStrObj.MatchString }

{************************************************************
 * TStrObj.Pos
 *
 * Parameters:
 *  aStr: pointer to a zero-terminated string to search for
 * Returns:
 *  The starting position of the found substring in self, counting
 *  from 0, or NOT_FOUND. if the substring was not found or the
 *  parameter was Nil or is an empty string.
 * Description:
 *  Searches for the occurence of the passed string ( case-sensitive ) 
 *  in the current string and returns the starting position. This 
 *  differs from the behaviour of StrPos, which returns a pointer.
 *  You can obtain a similar result using At( Pos( aStr )) but this
 *  should rarely be necessary since all the contents manipulation
 *  methods take positions as parameters.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:12:01 by P. Below
 ************************************************************}
Function TStrObj.Pos( aStr: PChar ): Word;
  Begin
    Pos := PosI( aStr, 0 );
  End; { TStrObj.Pos }

{************************************************************
 * TStrObj.PosI
 *
 * Parameters:
 *  aStr   : pointer to a zero-terminated string to search for
 *  startAt: position (counting from 0) in the current string
 *           to start searching at.
 * Returns:
 *  The starting position of the found substring in self, counting
 *  from 0, or NOT_FOUND. if the substring was not found or the
 *  parameter was Nil or is an empty string or startAt was beyond
 *. the end of string
 * Description:
 *  Searches for the occurence of the passed string ( case-sensitive ) 
 *  in the current string and returns the starting position. This 
 *  differs from the behaviour of StrPos, which returns a pointer.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:12:01 by P. Below
 ************************************************************}
Function TStrObj.PosI( aStr: PChar; startAt: Word ): Word;
  Var
    p: PChar;
  Begin
    PosI := NOT_FOUND;
    If IsValid and Assigned( aStr ) and ( StrLen( aStr ) > 0 )Then Begin
      If startAt >= GetLength Then
        Exit;
      p := Get+startAt;
      p := StrPos( p, aStr );
      If Assigned( p ) Then
        PosI := p - Get;
    End; { If }
  End; { TStrObj.PosI }

{************************************************************
 * TStrObj.PosString
 *
 * Parameters:
 *  aStr: a Pascal string to search for
 * Returns:
 *  The starting position of the found substring in self, counting
 *  from 1, or 0, if the substring was not found or is an empty string.
 * Description:
 *  Searches for the occurence of the passed string ( case-sensitive ) 
 *  in the current string and returns the starting position. This 
 *  is basically as the System.Pos function works but this one will
 *  search the whole of the current string, even if it is longer than
 *  255 bytes and can thus return values > 255!
 * Error Conditions:
 *  May run out of memory in the temp object used to convert the
 *  string to a pchar and will return 0 in this case.
 *
 *Created: 11/20/94 13:12:01 by P. Below
 ************************************************************}
Function TStrObj.PosString( Const aStr: String ): Word;
  Var
    tempObj: TStrObj;
  Begin
    tempObj.InitString( aStr );
    If not tempObj.Failed Then
      (* the following code depends on the feature that NOT_FOUND+1 = 0,
         so we need to switch overflow checking off here to prevent a
         run-time error in case the program is accidentially compiled
         with overflow checking on. *)
{$IFOPT Q+}
{$Q-}
{$DEFINE QWASON}
{$ENDIF}
      PosString := Pos( tempObj.Get ) + 1
{$IFDEF QWASON}
{$Q+}
{$UNDEF QWASON}
{$ENDIF}
    Else Begin
      ForwardError( tempObj );
      PosString := 0;
    End;
    tempObj.Done;
  End; { TStrObj.PosString }

{************************************************************
 * TStrObj.PosObj
 *
 * Parameters:
 *  aStrObj: a TStrObj 
 * Returns:
 *  The starting position of the found substring in self, counting
 *  from 0, or NOT_FOUND. if the substring was not found or the
 *  parameter was Nil or is an empty string.
 * Description:
 *  Searches for the occurence of the string in the passed object 
 *  ( case-sensitive ) in the current string and returns the starting 
 *  position. Uses Pos for the work.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:12:01 by P. Below
 ************************************************************}
Function TStrObj.PosObj( Const aStrObj: TStrObj ): Word;
  Begin
    PosObj := Pos( aStrObj.Get );
  End; { TStrObj.PosObj }


{************************************************************
 * TStrObj.At
 *
 * Parameters:
 *  index: position of the required character, counts from 0
 * Returns:
 *  a pointer to the requested character in our current string, 
 *  or nil, if the object is empty or index is beyond the end of 
 *  string. You can access the terminating #0 with index=GetLength.
 * Description:
 *  This method can be used to get a pointer for the substring of
 *  the objects string that starts at position index. This pointer
 *  should be used read-only! If you use it as a write target you may
 *  mess up the internal data and will bypass all inbuild safety 
 *  features!
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:52:10 by P. Below
 ************************************************************}
Function TStrObj.At( index: Word ): PChar;
  Begin
    At := Nil;
    If IsValid and ( index <= GetLength ) Then
      At := @Buf[ index+1 ];
  End; { TStrObj.At }

{************************************************************
 * TStrObj.AtChar
 *
 * Parameters:
 *  index: position of the required character, counts from 0
 * Returns:
 *  the character at the requested position or #0, if the object is
 *  empty or index is beyond the end of string.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:52:10 by P. Below
 ************************************************************}
Function TStrObj.AtChar( index: Word ): Char ;
  Begin
    AtChar := #0;
    If IsValid and ( index <= GetLength ) Then
      AtChar := Buf[ index+1 ];
  End; { TStrObj.AtChar }

{************************************************************
 * TStrObj.AtString
 *
 * Parameters:
 *  index: position of the required character, counts from 1.
 * Returns:
 *  the character at the requested position or #0, if the object is
 *  empty or index is beyond the end of string. Index is not limited
 *  to 1..255!
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 13:52:10 by P. Below
 ************************************************************}
Function TStrObj.AtString( index: Word ): Char;
  Begin
    If index > 0 Then Dec( index );
    AtString := AtChar( index );
  End; { TStrObj.AtString }

{************************************************************
 * TStrObj.Scan
 *
 * Parameters:
 *  c: character to search
 * Returns:
 *  The position of the first occurence of c in our current string
 *  ( counting from 0 ), or NOT_FOUND, if the character is not
 *  found or the object is empty.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.Scan( c: Char ): Word;
  Var
    p: PChar;
  Begin
    Scan := ScanI( c, 0 );
  End; { TStrObj.Scan }

{************************************************************
 * TStrObj.ScanI
 *
 * Parameters:
 *  c: character to search
 *  startAt: position ( counting from 0 ) to start scanning at.
 * Returns:
 *  The position of the first occurence of c in our current string
 *  ( counting from 0 ), or NOT_FOUND, if the character is not
 *  found or the object is empty or startAt is beyond the end 
 *  of string.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.ScanI( c: Char; startAt: Word ): Word;
  Var
    p: PChar;
  Begin
    ScanI := NOT_FOUND;
    If IsValid and ( startAt < GetLength )Then Begin
      p := StrScan( Get+startAt, c );
      If Assigned( p ) Then
        ScanI := p - Get;
    End;
  End; { TStrObj.ScanI }

{************************************************************
 * TStrObj.ScanString
 *
 * Parameters:
 *  c: character to search
 * Returns:
 *  The position of the first occurence of c in our current string
 *  ( counting from 1 ), or 0, if the character is not
 *  found or the object is empty.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.ScanString( c: Char ): Word;
  Begin
    (* the following code depends on the feature that NOT_FOUND+1 = 0,
       so we need to switch overflow checking off here to prevent a
       run-time error in case the program is accidentially compiled
       with overflow checking on. *)
{$IFOPT Q+}
{$Q-}
{$DEFINE QWASON}
{$ENDIF}
    ScanString := Scan( c )+1;
{$IFDEF QWASON}
{$Q+}
{$UNDEF QWASON}
{$ENDIF}
  End; { TStrObj.ScanString }

{************************************************************
 * TStrObj.RScan
 *
 * Parameters:
 *  c: character to search
 * Returns:
 *  The position of the last occurence of c in our current string
 *  ( counting from 0 ), or NOT_FOUND, if the character is not
 *  found or the object is empty.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.RScan( c: Char ): Word;
  Var
    p: PChar;
  Begin
    RScan := NOT_FOUND;
    If IsValid Then Begin
      p := StrRScan( Get, c );
      If Assigned( p ) Then
        RScan := p - Get;
    End
  End; { TStrObj.RScan }

{************************************************************
 * TStrObj.RScanI
 *
 * Parameters:
 *  c: character to search
 *  startAt: position ( counting from 0 ) to start scanning at.
 * Returns:
 *  The position of the last occurence of c at or before position
 *  startAt (counting from 0), or NOT_FOUND, if the character is not
 *  found or the object is empty. If startAt is > GetLength, the
 *  return is the same as RScan.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.RScanI( c: Char; startAt: Word ): Word;
  Var
    p: PChar;
  Begin
    RScanI := NOT_FOUND;
    If IsValid Then Begin
      p := Get;
      If startAt > GetLength Then
        startAt := GetLength;
      asm
        les di, p
        mov cx, startAt
        test cx, cx
        jz @not_found
        add di, cx
        std
        inc cx
        mov al, c
        repne scasb
        jne @not_found
        mov @result, cx
      @not_found:
        cld
      End;
    End; { If }
  End; { TStrObj.RScanI }

{************************************************************
 * TStrObj.RScanString
 *
 * Parameters:
 *  c: character to search
 * Returns:
 *  The position of the last occurence of c in our current string
 *  ( counting from 1 ), or 0, if the character is not
 *  found or the object is empty.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 14:58:48 by P. Below
 ************************************************************}
Function TStrObj.RScanString( c: Char ): Word;
  Begin
    (* the following code depends on the feature that NOT_FOUND+1 = 0,
       so we need to switch overflow checking off here to prevent a
       run-time error in case the program is accidentially compiled
       with overflow checking on. *)
{$IFOPT Q+}
{$Q-}
{$DEFINE QWASON}
{$ENDIF}
    RScanString := RScan( c )+1;
{$IFDEF QWASON}
{$Q+}
{$UNDEF QWASON}
{$ENDIF}
  End; { TStrObj.RScanString }

(* memory handling methods, considered for internal use only,
   only exposed for the benefit of derived objects! *)

{************************************************************
 * TStrObj.Alloc
 *
 * Parameters:
 *  newSize: required capacity of the memory to allocate.
 * Returns:
 *  true if memory could be allocated, false otherwise.
 * Description:
 *  The amount actually allocated will be SO_Extrabytes larger
 *  than the one required. Any already allocated memory will be
 *  be released first. The new memory is filled with #0.
 * Error Conditions:
 *  May run out of memory.
 *
 *Created: 11/20/94 15:01:55 by P. Below
 ************************************************************}
Function  TStrObj.Alloc( newSize: Word ): Boolean; 
  Type
    pWord = ^Word;
  Begin
    Alloc := False;
    DisposeMem;
    Inc( newSize, SO_Extrabytes );
    GetMem( Buf, newSize );
    If IsValid Then Begin
      size := newSize;
      FillChar( Buf^, newSize, 0 );
{$IFDEF DEBUG}
      pWord( Buf+newSize-2 )^:= SO_Guard;
{$ENDIF}
      Alloc := True;
    End
    Else
      Error( errOutOfMemory, newSize );
  End; { TStrObj.Alloc }

{************************************************************
 * TStrObj.DisposeMem
 *
 * Parameters:
 *  none
 * Description:
 *  Disposes of any allocated memory.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 15:05:43 by P. Below
 ************************************************************}
Procedure TStrObj.DisposeMem;
  Begin
    If IsValid Then
      FreeMem( Buf, size );
    size := 0;
  End; { TStrObj.DisposeMem }

{************************************************************
 * TStrObj.ReAlloc
 *
 * Parameters:
 *  newSize: new capacity to allocate for this object. This can
 *           be larger or smaller than the current size, or
 *           even equal, which is of cause a waste of CPU time.
 *  preseve: true, of the current contents of this object
 *           should be copied to the new memory, false if
 *           they can be ignored,
 * Description:
 *  Saves the current buffer pointer and size to local vars, then
 *  tries to allocate the new memory requested. If successful,
 *  it copies as much of the old data as will fit into the new
 *  space if preserve is true, and then disposes of the old memory.
 * Error Conditions:
 *  May run out of memory, in which case the old data is restored.
 *
 *Created: 11/20/94 15:06:12 by P. Below
 ************************************************************}
Procedure TStrObj.ReAlloc( newsize: Word; preserve: Boolean );
  Var
    tempP: Pointer;
    tempS: Word;
    n    : Word;
  Begin
    tempP := Buf;
    tempS := size;
    Buf := Nil;
    If not Alloc( newsize ) Then Begin
      (* failed, restore old data *)
      Buf := tempP;
      size := tempS;
    End { If }
    Else Begin
      If Assigned( tempP ) Then Begin
        If preserve Then Begin
          If tempS > size Then
            n := size
          Else
            n := tempS;
  {$IFDEF DEBUG}
            Dec( n, 2 ); (* don't copy the guard word *)
  {$ENDIF}
          Move( tempP^, Buf^, n );
        End; { If }
        FreeMem( tempP, tempS );
      End; { If }
    End; { Else }
  End; { TStrObj.ReAlloc }

{************************************************************
 * TStrObj.Compact
 *
 * Parameters:
 *  none
 * Description:
 *  This method will try to reduce the allocated memory for this
 *  objects data to the size actually needed. This may fail, 
 *  because the new space is allocated before the old space is
 *  released! However, data is never lost.
 * Error Conditions:
 *  May run out of memory.
 *
 *Created: 11/20/94 15:13:48 by P. Below
 ************************************************************}
Procedure TStrObj.Compact;
  Var
    needed: Word;
  Begin
    needed := GetLength;
    If GetCapacity > needed Then
      ReAlloc( needed, True );
  End; { TStrObj.Compact }

{************************************************************
 * TStrObj.SetLength
 *
 * Parameters:
 *  len: length of the current Pchar
 * Description:
 *  This methods sets the length byte for the Pascal-string 
 *  representation of our data. If the passed len is < 256, 
 *  is directly yields the length byte, for all larger length
 *  the pascal string length is 255.
 * Error Conditions:
 *  none
 *
 *Created: 11/20/94 15:15:26 by P. Below
 ************************************************************}
Procedure TStrObj.SetLength( len: Word );
  Begin
    If len > 255 Then
      Buf^ := #255
    Else
      Buf^ := Chr( len )
  End; (* TStrObj.SetLength *)

{************************************************************
 * TStrObj.CalcLength
 *
 * Parameters:
 *  none
 * Description:
 *  Recalculates the length byte for the current string. Use
 *  this procedure if you pass the return of Get to some procedure
 *  that writes to the buffer. Use this approach with caution 
 *  because the error checks and automatic memory handling of 
 *  TStrObj are bypassed!
 * Error Conditions:
 *  none
 *
 *Created: 01/28/95 15:18:42 by P. Below
 ************************************************************}
Procedure TStrObj.CalcLength;
  Begin
    If IsValid Then
      SetLength( StrLen( Get ));
  End; (* TStrObj.CalcLength *)

{************************************************************
 * TStrObj.Fits
 *
 * Parameters:
 *  neededSize: capacity needed for the data to be put into the
 *              object.
 *  preserve  : wether to preserve the old contents of the object
 *              if the memory needs to grow.
 * Returns:
 *  true if the object can hold neededSize characters, either because
 *  it had enough memory allocated to begin with or because it has been
 *  successfully grown, false otherwise.
 * Description:
 *  Checks if the current capacity of the object is >= neededSize, and
 *  if not, tries to grow to the requested size.
 * Error Conditions:
 *  May run out of memory on attempt to grow the allocated memory block.
 *
 *Created: 11/20/94 15:17:15 by P. Below
 ************************************************************}
Function  TStrObj.Fits( neededSize: TLen; preserve: Boolean ): Boolean;
  Begin
    If neededSize >= ( $10000 - SO_Extrabytes) Then Begin
      Error( SOERR_BlockTooLarge, 0 );
      Fits := False;
    End
    Else
      If LoWord( neededSize ) >= GetCapacity Then Begin
        ReAlloc( LoWord( neededSize ), preserve );
        Fits := not Failed;
      End { If }
      Else
        Fits := True;
  End; { TStrObj.Fits }

{+------------------------------------------------------------------ 
 | Methods of TFStrObj
 +------------------------------------------------------------------}
Constructor TFStrObj.InitReadLn( Var f: Text );
  Var
    err: Integer;
  Begin
    InitSize( 255 );
    If not Failed Then Begin
{$I-}
      ReadLn( f, GetPString^ );
{$I+}
      err := IOresult;
      If err <> 0 Then Begin
        Error( errFilesystem, err );
        DisposeMem;
      End
      Else
        Compact;
    End; { If }
  End; { TFStrObj.InitReadLn }

Constructor TFStrObj.InitRead( Var f: Text; count: Word );
  Var
    err, i: Integer;
    p: Pchar;
  Begin
    InitSize( count );
    If not Failed Then Begin
      err := 0;
      i   := 0;
      p   := Get;
{$I-}
      While not Eof( f ) and ( err=0 ) and ( i < count ) Do Begin
        Read( f, p^ );
        err := IOresult;
        If err = 0 Then Begin
          Inc( i );
          Inc( p );
        End
        Else
          Error( errFilesystem, err );
      End; { While }
{$I+}
      SetLength( i );
    End; { If }
  End; { TFStrObj.InitRead }

Procedure TFStrObj.FWriteLn( Var f: Text );
  Var
    err: Integer;
  Begin
    If Isvalid Then Begin
{$I-}
      WriteLn( f, Get );
      err := IOresult;
      If err <> 0 Then
        Error( errFilesystem, err );
{$I+}
    End; { If }
  End; { TFStrObj.FWriteLn }

Procedure TFStrObj.FWrite( Var f: Text );
  Var
    err: Integer;
  Begin
    If Isvalid Then Begin
{$I-}
      Write( f, Get );
      err := IOresult;
      If err <> 0 Then
        Error( errFilesystem, err );
{$I+}
    End; { If }
  End; { TFStrObj.FWrite }

(* NewObject provides an empty but initialized object, freshly
   allocated on the heap. This method has to be replaced by any
   derived objects to return an instance of their own object type! *)
Function TFStrObj.NewObject: PStrObj;
  Begin
    NewObject := New( PFStrObj, InitBlank );
  End; { TFStrObj.NewObject }

{+------------------------------------------------------------------
 | Methods of TMsgObj
 +------------------------------------------------------------------}
Function TMsgObj.Display( parentHwnd: HWnd; title: PChar; 
    msgboxFlags: Word ): Word;
  Begin
    Display := IDCANCEL;
    If IsValid Then
{$IFDEF WINDOWS}
      Display := MessageBox( parentHWnd, Get, title, msgboxFlags )
{$ENDIF}
  End; { TMsgObj.Display }

Function TMsgObj.DisplayWithID( parentHwnd: HWnd; resID, 
    msgboxFlags: Word ): Word;
  Var
    tempObj: TStrObj;
  Begin
    DisplayWithID := IDCANCEL;
    tempObj.InitResourceString( HInstance, resID );
    If not tempObj.Failed Then 
      DisplayWithID := Display( parentHwnd, tempObj.Get, msgboxFlags )
    Else
      ForwardError( tempObj );
    tempObj.Done;
  End; { TMsgObj.DisplayWithID }

(* NewObject provides an empty but initialized object, freshly
   allocated on the heap. This method has to be replaced by any
   derived objects to return an instance of their own object type! *)
Function TMsgObj.NewObject: PStrObj;
  Begin
    NewObject := New( PMsgObj, InitBlank );
  End; { TStrObj.NewObject }

(* no unit initialization needed *)
End.
